import unittest, unittest.mock
from unittest.mock import call
from collections.abc import Sequence

from qmi.instruments.thorlabs import Thorlabs_Tc200
from qmi.instruments.thorlabs.tc200 import *
from qmi.instruments.thorlabs.tc200 import _Query, _Command

import qmi.core.exceptions
from qmi.core.context import QMI_Context


class TestThorlabsTc200(unittest.TestCase):

    def setUp(self) -> None:
        unittest.mock.patch("qmi.core.transport.QMI_SerialTransport._validate_device_name")
        qmi_context = unittest.mock.MagicMock(spec=QMI_Context)
        qmi_context.name = "mockytemp"
        self.ser_address = "COM298"
        self.baudrate = 115200
        transport_id = "serial:{}".format(self.ser_address)
        self.thorlabs = Thorlabs_Tc200(qmi_context, "heet", transport_id)

    def test_open_close(self):
        with unittest.mock.patch("serial.Serial") as ser:
            self.thorlabs.open()
            self.thorlabs.close()

            ser.assert_called_once_with(
                self.ser_address,
                baudrate=self.baudrate,  # The rest are defaults
                bytesize=8,
                parity='N',
                rtscts=False,
                stopbits=1.0,
                timeout=0.04
                )

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_receive_response(self, mock_read):
        """See that the _receive_response method receives one block of data"""
        mock_read.return_value = bytearray("query_cmd?\rblablabla\r> ".encode('ascii'))
        expected = "blablabla"
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            msg = self.thorlabs._receive_response()
            self.thorlabs.close()

        self.assertEqual(msg, expected)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_receive_response_with_parts(self, mock_read):
        """See that _receive_response method receives multiple parts of data"""
        mock_read.return_value = bytearray("fake_query?\rpart1\rpart2\rpart3\r> ".encode('ascii'))
        expected = "part1\npart2\npart3"
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            msg = self.thorlabs._receive_response()
            self.thorlabs.close()

        self.assertEqual(msg, expected)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_receive_response_raises_unicodeerror(self, mock_read):
        """"An error should be raised if the extra data present is not of expected length"""
        mock_read.return_value = bytearray("fake_query?\r".encode('ascii') + b"\x81 > ")
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs._receive_response()

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_query_response_raises_qmi_instrumentexception(self, mock_read):
        """"An error should be raised if the extra data present is not of expected length"""
        mock_read.return_value = bytearray(f"query_cmd?\rCMD_NOT_DEFINED\r> ".encode('ascii'))
        query_mock = unittest.mock.Mock(spec=_Query)
        query_mock.value = "Nonsense"
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs._send_query(query_mock)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_idn(self, mock_read):
        # arrange
        expected_call = _Query.IDENT.value + "\r"
        vendor = "Thorlabs"
        model_number = "567890"
        # no serial number
        smth_else = "not_needed"
        fw_version = "12.3.4"
        input_string = expected_call + " ".join([vendor, model_number, smth_else, fw_version]) + "\r> "
        expected = 'QMI_InstrumentIdentification(vendor=\'{}\', model=\'{}\', serial=\'{}\', version=\'{}\')'.format(
            vendor, model_number, "", fw_version)
        mock_read.return_value = bytearray(input_string.encode())
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            idn = self.thorlabs.get_idn()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        # assert
        self.assertEqual(str(idn), expected)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_status(self, mock_read):
        """Test that get_status() returns a set of "random" status bits."""
        # arrange
        expected_call = _Query.STATUS.value + "\r"
        expected_response = Tc200Status(
            output_state=OutputState(1),
            mode=Mode(0),
            sensor_type=SensorType(8),
            unit=DisplayUnit(0),
            alarm=SensorAlarmState(64),
            cycle_state=CycleState(0)
        )
        mock_read.return_value = bytearray(str(expected_call + "149 > ").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            status = self.thorlabs.get_status()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        # assert
        self.assertEqual(status, expected_response)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_status_raises_valueerror(self, mock_read):
        """Test that get_status() raises ValueError on invalid response."""
        # arrange
        status_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray("{}iob > ".format(status_call).encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
                self.thorlabs.get_status()
                pass

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_status_raises_unicodeerror(self, mock_read):
        """Test that get_status() raises ValueError on invalid response."""
        # arrange
        status_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray(status_call.encode() + b"\x81 > ")
        # act
        with unittest.mock.patch("serial.Serial"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.get_status()

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_is_enabled(self, mock_read):
        """Test that get_enabled() returns expected state."""
        # arrange
        expected_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "149 > ").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            status = self.thorlabs.is_enabled()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        # assert
        self.assertTrue(status)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_has_alarm(self, mock_read):
        """Test that has_alarm() returns expected state."""
        # arrange
        expected_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "149 > ").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            status = self.thorlabs.has_alarm()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        # assert
        self.assertTrue(status)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_is_tuned(self, mock_read):
        """Test that is_tuned() returns expected state (True)."""
        # arrange
        expected_call = _Query.TUNED.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "1\r > ").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            status = self.thorlabs.is_tuned()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        # assert
        self.assertTrue(status)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_enable(self, mock_read):
        """Test that enable() enables the heater output."""
        # arrange
        expected_calls = [
            _Query.STATUS.value + "\r",
            _Query.STATUS.value + "\r",
            _Command.TOGGLE_ENABLE.value + "\r"
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "1 > \r").encode("ascii")),
            bytearray(str(expected_calls[1] + "0 > \r").encode("ascii")),
            bytearray(str(expected_calls[2] + "OK > \r").encode("ascii"))
        ]
        expected_sequence: Sequence = map(call, map(str.encode, expected_calls))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.enable()
            self.thorlabs.close()
            wrt.assert_has_calls(expected_sequence, any_order=True)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_enable_not_needed(self, mock_read):
        """Test that enable() does not call TOGGLE_ENABLE if already enabled."""
        # arrange
        expected_calls = [
            _Query.STATUS.value + "\r",
            _Query.STATUS.value + "\r",
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "1 > \r").encode("ascii")),
            bytearray(str(expected_calls[1] + "1 > \r").encode("ascii")),
        ]
        expected_sequence: Sequence = map(call, map(str.encode, expected_calls))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.enable()
            self.thorlabs.close()
            wrt.assert_has_calls(expected_sequence, any_order=True)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_enable_excepts(self, mock_read):
        """Test that enable() excepts if it has alarm."""
        # arrange
        expected_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "149 > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.enable()

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_disable(self, mock_read):
        """Test that disable() disables the heater output."""
        # arrange
        expected_calls = [
            _Query.STATUS.value + "\r",
            _Command.TOGGLE_ENABLE.value + "\r"
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "1 > \r").encode("ascii")),
            bytearray(str(expected_calls[1] + "OK > \r").encode("ascii"))
        ]
        expected_sequence: Sequence = map(call, map(str.encode, expected_calls))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.disable()
            self.thorlabs.close()
            wrt.assert_has_calls(expected_sequence, any_order=True)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_disable_not_needed(self, mock_read):
        """Test that disable() does not call TOGGLE_ENABLE if already disabled."""
        # arrange
        expected_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "0 > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.disable()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_temperature(self, mock_read):
        """Test that set_temperature() command is sent."""
        # arrange
        temperature = 25.0
        expected_call = _Command.SET_TEMPERATURE.value + f"={temperature:.1f}\r"
        mock_read.return_value = bytearray(str(expected_call + "OK > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_temperature(temperature)
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_temperature_out_of_range(self, mock_read):
        """Test that set_temperature() raises exceptions when T is out of range."""
        # arrange
        temperatures = [TMIN - 1, TMAX + 1]
        set_calls = [_Command.SET_TEMPERATURE.value + f"={temperatures[0]:.1f}\r",
                     _Command.SET_TEMPERATURE.value + f"={temperatures[1]:.1f}\r"
        ]
        for t, set_call in enumerate(set_calls):
            mock_read.return_value = bytearray(str(set_call + "OK > \r").encode("ascii"))
            # act
            with unittest.mock.patch("serial.Serial"), \
                    unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
                self.thorlabs.open()
                with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
                    self.thorlabs.set_temperature(temperatures[t])

                self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_temperature_hardware_error(self, mock_read):
        """Test that set_temperature() command catches a hardware error."""
        # arrange
        temperature = 25.0
        mock_read.return_value = bytearray((_Command.SET_TEMPERATURE.value +
                                            "\rCMD_ARG_RANGE_ERR > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.set_temperature(temperature)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_temperature_setpoint(self, mock_read):
        """Test that get_temperature_setpoint() obtains expected T setpoint."""
        # arrange
        temperature = 25.0
        expected_call = _Query.SETPOINT_TEMPERATURE.value + "\r"
        mock_read.return_value = bytearray((expected_call + f"{temperature:.1f} Celsius\r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            response = self.thorlabs.get_temperature_setpoint()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        self.assertEqual(response, temperature)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_temperature_actual(self, mock_read):
        """Test that get_temperature_setpoint() obtains expected T setpoint."""
        # arrange
        temperature = 25.0
        expected_call = _Query.ACTUAL_TEMPERATURE.value + "\r"
        mock_read.return_value = bytearray((expected_call + f"{temperature:.1f} C\r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            response = self.thorlabs.get_temperature_actual()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        self.assertEqual(response, temperature)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_temperature_limit(self, mock_read):
        """Test that set_temperature_limit() command is sent."""
        # arrange
        temperature = 25.0
        expected_call = _Command.MAX_TEMPERATURE.value + f"={temperature:.1f}\r"
        mock_read.return_value = bytearray(str(expected_call + "OK > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_temperature_limit(temperature)
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_temperature_limit_out_of_range(self, mock_read):
        """Test that set_temperature_limit() raises exceptions when T is out of range."""
        # arrange
        temperatures = [TLIMIT_MIN - 1, TLIMIT_MAX + 1]
        set_calls = [_Command.MAX_TEMPERATURE.value + f"={temperatures[0]:.1f}\r",
                     _Command.MAX_TEMPERATURE.value + f"={temperatures[1]:.1f}\r"
        ]
        for t, set_call in enumerate(set_calls):
            mock_read.return_value = bytearray(str(set_call + "OK > \r").encode("ascii"))
            # act
            with unittest.mock.patch("serial.Serial"), \
                    unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
                self.thorlabs.open()
                with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
                    self.thorlabs.set_temperature_limit(temperatures[t])

                self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_temperature_limit_hardware_error(self, mock_read):
        """Test that set_temperature_limit() command catches a hardware error."""
        # arrange
        temperature = 25.0
        mock_read.return_value = bytearray((_Command.MAX_TEMPERATURE.value +
                                            "\rCMD_ARG_RANGE_ERR > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.set_temperature_limit(temperature)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_temperature_limit(self, mock_read):
        """Test that get_temperature_limit() obtains expected T limit."""
        # arrange
        temperature = 25.0
        expected_call = _Query.MAX_TEMPERATURE.value + "\r"
        mock_read.return_value = bytearray((expected_call + f"{temperature:.1f}\r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            response = self.thorlabs.get_temperature_limit()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        self.assertEqual(response, temperature)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_power_limit(self, mock_read):
        """Test that set_power_limit() command is sent."""
        # arrange
        power = 10.0
        expected_call = _Command.MAX_POWER.value + f"={power:.1f}\r"
        mock_read.return_value = bytearray(str(expected_call + "OK > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_power_limit(power)
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_power_limit_out_of_range(self, mock_read):
        """Test that set_power_limit() raises exceptions when P is out of range."""
        # arrange
        powers = [PLIMIT_MIN - 1, PLIMIT_MAX + 1]
        set_calls = [_Command.MAX_POWER.value + f"={powers[0]:.1f}\r",
                     _Command.MAX_POWER.value + f"={powers[1]:.1f}\r"
        ]
        for t, set_call in enumerate(set_calls):
            mock_read.return_value = bytearray(str(set_call + "OK > \r").encode("ascii"))
            # act
            with unittest.mock.patch("serial.Serial"), \
                    unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
                self.thorlabs.open()
                with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
                    self.thorlabs.set_power_limit(powers[t])

                self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_power_limit_hardware_error(self, mock_read):
        """Test that set_power_limit() command catches a hardware error."""
        # arrange
        power = 10.0
        mock_read.return_value = bytearray((_Command.MAX_POWER.value +
                                            "\rCMD_ARG_RANGE_ERR > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.set_power_limit(power)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_power_limit(self, mock_read):
        """Test that get_power_limit() obtains expected P limit."""
        # arrange
        power = 12.0
        expected_call = _Query.MAX_POWER.value + "\r"
        mock_read.return_value = bytearray((expected_call + f"{power:.1f}\r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            response = self.thorlabs.get_power_limit()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        self.assertEqual(response, power)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_p(self, mock_read):
        """Test that set_pid_gains(...) command is sent and sets proportional gain."""
        # arrange
        proportion = 10
        expected_call = _Command.PGAIN.value + f"={proportion}\r"
        mock_read.return_value = bytearray(str(expected_call + "OK > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_pid_gains(proportion, None, None)
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_i(self, mock_read):
        """Test that set_pid_gains(...) command is sent and sets integralal gain."""
        # arrange
        integral = 10
        expected_call = _Command.IGAIN.value + f"={integral}\r"
        mock_read.return_value = bytearray(str(expected_call + "OK > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_pid_gains(None, integral, None)
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_d(self, mock_read):
        """Test that set_pid_gains(...) command is sent and sets differentional gain."""
        # arrange
        differention = 10
        expected_call = _Command.DGAIN.value + f"={differention}\r"
        mock_read.return_value = bytearray(str(expected_call + "OK > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_pid_gains(None, None, differention)
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_all(self, mock_read):
        """Test that set_pid_gains(...) command is sent and sets all gains."""
        # arrange
        proportion = 10
        integral = 10
        differention = 10
        expected_calls = [_Command.PGAIN.value + f"={proportion}\r",
                          _Command.IGAIN.value + f"={integral}\r",
                          _Command.DGAIN.value + f"={differention}\r"
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "OK > \r").encode("ascii")),
            bytearray(str(expected_calls[1] + "OK > \r").encode("ascii")),
            bytearray(str(expected_calls[2] + "OK > \r").encode("ascii"))
        ]
        expected_sequence: Sequence = map(call, map(str.encode, expected_calls))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.set_pid_gains(proportion, integral, differention)
            self.thorlabs.close()
            wrt.assert_has_calls(expected_sequence)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_out_of_range_for_p(self, mock_read):
        """Test that set_pid_gains(p, ...) raises exceptions when P is out of range."""
        # arrange
        proportionals = [PGAIN_MIN - 1, PGAIN_MAX + 1]
        set_calls = [_Command.PGAIN.value + f"={proportionals[0]}\r",
                     _Command.PGAIN.value + f"={proportionals[1]}\r"
        ]
        for t, set_call in enumerate(set_calls):
            mock_read.return_value = bytearray(str(set_call + "OK > \r").encode("ascii"))
            # act
            with unittest.mock.patch("serial.Serial"), \
                    unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
                self.thorlabs.open()
                with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
                    self.thorlabs.set_pid_gains(proportionals[t], None, None)

                self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_for_p_hardware_error(self, mock_read):
        """Test that set_pid_gains(p, ...) command catches a hardware error."""
        # arrange
        proportional = 10
        mock_read.return_value = bytearray((_Command.PGAIN.value +
                                            "\rCMD_ARG_RANGE_ERR > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.set_pid_gains(proportional, None, None)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_out_of_range_for_i(self, mock_read):
        """Test that set_pid_gains(..., i, ...) raises exceptions when I is out of range."""
        # arrange
        integrals = [IGAIN_MIN - 1, IGAIN_MAX + 1]
        set_calls = [_Command.IGAIN.value + f"={integrals[0]}\r",
                     _Command.IGAIN.value + f"={integrals[1]}\r"
        ]
        for t, set_call in enumerate(set_calls):
            mock_read.return_value = bytearray(str(set_call + "OK > \r").encode("ascii"))
            # act
            with unittest.mock.patch("serial.Serial"), \
                    unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
                self.thorlabs.open()
                with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
                    self.thorlabs.set_pid_gains(None, integrals[t], None)

                self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_for_i_hardware_error(self, mock_read):
        """Test that set_pid_gains(..., i, ...) command catches a hardware error."""
        # arrange
        integral = 10
        mock_read.return_value = bytearray((_Command.IGAIN.value +
                                            "\rCMD_ARG_RANGE_ERR > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.set_pid_gains(None, integral, None)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_out_of_range_for_d(self, mock_read):
        """Test that set_pid_gains(..., d) raises exceptions when D is out of range."""
        # arrange
        differentionals = [DGAIN_MIN - 1, DGAIN_MAX + 1]
        set_calls = [_Command.DGAIN.value + f"={differentionals[0]}\r",
                     _Command.DGAIN.value + f"={differentionals[1]}\r"
        ]
        for t, set_call in enumerate(set_calls):
            mock_read.return_value = bytearray(str(set_call + "OK > \r").encode("ascii"))
            # act
            with unittest.mock.patch("serial.Serial"), \
                    unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
                self.thorlabs.open()
                with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
                    self.thorlabs.set_pid_gains(None, None, differentionals[t])

                self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_set_pid_gains_for_d_hardware_error(self, mock_read):
        """Test that set_pid_gains(d, ...) command catches a hardware error."""
        # arrange
        differentional = 10
        mock_read.return_value = bytearray((_Command.DGAIN.value +
                                            "\rCMD_ARG_RANGE_ERR > \r").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write"):
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
                self.thorlabs.set_pid_gains(None, None, differentional)

            self.thorlabs.close()

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_get_pid_gains(self, mock_read):
        """Test that get_pid_gains() obtains expected P, I, D values."""
        # arrange
        proportional, integral, differential = 12, 11, 13
        expected_call = _Query.PID_GAINS.value + "\r"
        mock_read.return_value = bytearray(
            (expected_call + f"{proportional} {integral} {differential}\r").encode("ascii")
        )
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            p_gain, i_gain, d_gain = self.thorlabs.get_pid_gains()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())

        self.assertEqual(p_gain, proportional)
        self.assertEqual(i_gain, integral)
        self.assertEqual(d_gain, differential)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_tune(self, mock_read):
        """Test that tune() tunes the temperature offset."""
        # arrange
        expected_calls = [
            _Query.STATUS.value + "\r",
            _Query.TUNED.value + "\r",
            _Command.TUNE_OFFSET.value + "\r"
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "1 > \r").encode("ascii")),
            bytearray(str(expected_calls[1] + "0 \r> ").encode("ascii")),
            bytearray(str(expected_calls[2] + "OK > \r").encode("ascii"))
        ]
        expected_sequence: Sequence = map(call, map(str.encode, expected_calls))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.tune()
            self.thorlabs.close()
            wrt.assert_has_calls(expected_sequence, any_order=True)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_tune_excepts_alread_tuned(self, mock_read):
        """Test that tune() does excepts if already tuned."""
        # arrange
        expected_calls = [
            _Query.STATUS.value + "\r",
            _Query.TUNED.value + "\r",
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "1 > \r").encode("ascii")),
            bytearray(str(expected_calls[1] + "1 \r> ").encode("ascii")),
        ]
        expected_error = "Device is already tuned"
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_UsageException) as exc:
                self.thorlabs.tune()

            self.thorlabs.close()
            self.assertEqual(str(exc.exception), expected_error)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_tune_excepts_if_heater_not_enabled(self, mock_read):
        """Test that tune() excepts if heater is not enabled."""
        # arrange
        expected_call = _Query.STATUS.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "0 \r> ").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            with self.assertRaises(qmi.core.exceptions.QMI_UsageException) as exc:
                self.thorlabs.tune()

            self.thorlabs.close()
            self.assertEqual(str(exc.exception), "Heater output must be enabled for tuning")

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_detune(self, mock_read):
        """Test that detune() detunes the temperature offset."""
        # arrange
        expected_calls = [
            _Query.TUNED.value + "\r",
            _Command.TUNE_OFFSET.value + "\r"
        ]
        mock_read.side_effect = [
            bytearray(str(expected_calls[0] + "1\r> ").encode("ascii")),
            bytearray(str(expected_calls[1] + "OK\r> ").encode("ascii"))
        ]
        expected_sequence: Sequence = map(call, map(str.encode, expected_calls))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.detune()
            self.thorlabs.close()
            wrt.assert_has_calls(expected_sequence, any_order=True)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read_until")
    def test_detune_not_needed(self, mock_read):
        """Test that detune() does not call TUNE_OFFSET if already detuned."""
        # arrange
        expected_call = _Query.TUNED.value + "\r"
        mock_read.return_value = bytearray(str(expected_call + "0 \r> ").encode("ascii"))
        # act
        with unittest.mock.patch("serial.Serial"), \
                unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.write") as wrt:
            self.thorlabs.open()
            self.thorlabs.detune()
            self.thorlabs.close()
            wrt.assert_called_once_with(expected_call.encode())


if __name__ == '__main__':
    unittest.main()



import logging
from unittest import TestCase
from unittest.mock import patch, Mock, call

import numpy.testing
from numpy import nan

from qmi.core.exceptions import QMI_InvalidOperationException, QMI_InstrumentException
from qmi.instruments.high_finesse import HighFinesse_Wlm
from qmi.instruments.high_finesse.support import (
    wlmData,
    WlmGetErr,
)
from tests.patcher import QMI_Context

# Disable all logging
logging.disable(logging.CRITICAL)


@patch("qmi.instruments.high_finesse.support._library_wrapper.wlmData.LoadDLL")
class TestHighFinesse_Wlm(TestCase):
    """Test the behavior of the High Finesse driver (with the actual dll mocked)."""
    def setUp(self):
        # Start QMI context
        self.ctx = QMI_Context("highfinesse_wlm_client")

    def test_open(self, dll_patcher):
        """The connection should be 'open' after open() was (implicitly) called."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertTrue(wlm.is_open())

        dll_patcher.assert_called_once()

    def test_open_error(self, dll_patcher):
        """The driver should give an error when an open connection is opened again."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InvalidOperationException) as exc:
                wlm.open()

        self.assertEqual("Operation not allowed on open instrument wlm", str(exc.exception))

    def test_close(self, dll_patcher):
        """The connection should not be open after close() is called."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount = {0: 0}.get

        wlm = HighFinesse_Wlm(Mock(), "support")
        self.assertFalse(wlm.is_open())
        wlm.open()
        self.assertTrue(wlm.is_open())
        wlm.close()
        self.assertFalse(wlm.is_open())

    def test_close_error(self, dll_patcher):
        """Calling close() is not allowed when open() has not been called (implicitly)."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount = {0: 0}.get
        wlm = HighFinesse_Wlm(Mock(), "support")

        with self.assertRaises(QMI_InvalidOperationException) as exc:
            wlm.close()

        self.assertEqual("Operation not allowed on closed instrument support", str(exc.exception))

    def test_get_version(self, dll_patcher):
        """The driver should give the version string in the right format."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetWLMVersion.side_effect = {0: 1, 1: 12, 2: 0, 3: 20.4}.get

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertEqual(wlm.get_version(), "WLM Version: [1.12.0.20.4]")

        wlm._lib.dll.GetWLMVersion.assert_has_calls([call(0), call(1), call(2), call(3)])

    def test_get_idn(self, dll_patcher):
        """The driver should give the version string in the right format."""
        model = 6
        version = 123
        revision = 4321
        sw_ver = f"{version}.{revision}"
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetWLMVersion.side_effect = {0: model, 1: version, 2: revision, 3: 987.654}.get

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            idn = wlm.get_idn()
            self.assertEqual(idn.vendor, "HighFinesse")
            self.assertEqual(idn.model, f"WLM-{model}")
            self.assertIsNone(idn.serial)
            self.assertEqual(idn.version, sw_ver)

        wlm._lib.dll.GetWLMVersion.assert_has_calls([call(0), call(1), call(2), call(3)])

    def test_get_operation_state(self, dll_patcher):
        """Test a happy flow with getting an operation state."""
        expected_op_state = 0  # stopped
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetOperationState.return_value = expected_op_state

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertEqual(expected_op_state, wlm.get_operation_state())

        wlm._lib.dll.GetOperationState.assert_called_once_with(expected_op_state)

    def test_get_frequency(self, dll_patcher):
        """The driver should return the frequency value that the library outputs."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetFrequencyNum.return_value = 42.

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertEqual(42., wlm.get_frequency(1))

        wlm._lib.dll.GetFrequencyNum.assert_called_once_with(1, 0.0)

    def test_get_frequency_error(self, dll_patcher):
        """The get_frequency() method should raise an error when the library gave an error code."""
        expected_error = 0.  # The call returns a double
        enum_name = f"WlmGetErr.{WlmGetErr(expected_error).name}"
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetFrequencyNum.return_value = expected_error

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InstrumentException) as exc:
                wlm.get_frequency(1)

        self.assertEqual(f"Error received from library call 'GetFrequencyNum': {enum_name}", str(exc.exception))

    def test_get_frequency_error_unknown(self, dll_patcher):
        """The get_frequency() method should raise an error when the library gave an error code."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetFrequencyNum.return_value = -42

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InstrumentException) as exc:
                wlm.get_frequency(1)

        self.assertEqual("Error received from library call 'GetFrequencyNum': -42", str(exc.exception))

    def test_get_wavelength(self, dll_patcher):
        """The driver should return the wavelength value that the library outputs."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetWavelengthNum.return_value = 42.

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertEqual(42., wlm.get_wavelength(1))

        wlm._lib.dll.GetWavelengthNum.assert_called_once_with(1, 0.0)

    def test_get_wavelength_error(self, dll_patcher):
        """The get_wavelength() method should raise an error when the library gave an error code."""
        expected_error = -5.  # The call returns a double
        enum_name = f"WlmGetErr.{WlmGetErr(expected_error).name}"
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetWavelengthNum.return_value = expected_error

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InstrumentException) as exc:
                wlm.get_wavelength(1)

        self.assertEqual(f"Error received from library call 'GetWavelengthNum': {enum_name}", str(exc.exception))

    def test_get_wavelength_channel_error(self, dll_patcher):
        """The get_wavelength() method should raise an error when the library gave an error code."""
        invalid_channel = 9
        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InstrumentException) as exc:
                wlm.get_wavelength(invalid_channel)

        self.assertEqual(f"Channel number out of range: {invalid_channel}", str(exc.exception))

    def test_get_power(self, dll_patcher):
        """The driver should return the power value that the library outputs."""
        expected_power = 1.23
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetPowerNum.return_value = expected_power

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertEqual(expected_power, wlm.get_power(1))

        wlm._lib.dll.GetPowerNum.assert_called_once_with(1, 0.0)

    def test_get_power_error(self, dll_patcher):
        """The get_power() method should raise an error when the library gave an error code."""
        expected_error = -2.  # The call returns a double
        enum_name = f"WlmGetErr.{WlmGetErr(expected_error).name}"
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetPowerNum.return_value = expected_error

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InstrumentException) as exc:
                wlm.get_power(1)

        self.assertEqual(f"Error received from library call 'GetPowerNum': {enum_name}", str(exc.exception))

    def test_get_power_returns_nan(self, dll_patcher):
        """The get_power() method allows NaN as a valid return value."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetPowerNum.return_value = nan

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
             power = wlm.get_power(1)

        numpy.testing.assert_equal(nan, power)

    def test_set_data_pattern_defaults(self, dll_patcher):
        """Test the set_data_patterns calls SetPattern with default index and i_enable."""
        default_index, default_enable = 1, True
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.SetPattern.return_value = 1  # no errors

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertIsNone(wlm.set_data_pattern())

        wlm._lib.dll.SetPattern.assert_called_once_with(default_index, default_enable)

    def test_set_data_pattern_inputs(self, dll_patcher):
        """Test the set_data_patterns calls SetPattern with correct index and i_enable."""
        expected_index, expected_enable = 4, False
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.SetPattern.return_value = 1  # no errors

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            self.assertIsNone(wlm.set_data_pattern(expected_index, expected_enable))

        wlm._lib.dll.SetPattern.assert_called_once_with(expected_index, expected_enable)

    def test_set_data_pattern_invalid_index(self, dll_patcher):
        """Test the set_data_patterns calls SetPattern with correct index and i_enable."""
        invalid_index = 10
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(ValueError):
                wlm.set_data_pattern(invalid_index)

    def test_get_data_pattern_data(self, dll_patcher):
        """Test get_data_pattern_data for getting data."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetPatternDataNum.return_value = 1  # exporting was enabled

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            data = wlm.get_data_pattern_data(1, 1, 2048)

        self.assertEqual(2048, data.size)
        wlm._lib.dll.GetPatternDataNum.assert_called_once()

    def test_get_data_pattern_data_invalid_index(self, dll_patcher):
        """Test get_data_pattern_data excepts with invalid index number."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(ValueError):
                wlm.get_data_pattern_data(1, 6, 2048)

    def test_get_data_pattern_data_pattern_not_enabled(self, dll_patcher):
        """Test get_data_pattern_data when the channel/index was not enabled."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetPatternDataNum.return_value = 0  # exporting was disabled

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            data = wlm.get_data_pattern_data(1, 1, 2048)

        self.assertEqual(2048, data.size)
        wlm._lib.dll.GetPatternDataNum.assert_called_once()

    def test_get_data_pattern_data_error(self, dll_patcher):
        """Test get_data_pattern_data excepts with error code."""
        wlmData.dll = dll_patcher.return_value
        wlmData.dll.GetWLMCount.side_effect = {0: 1}.get
        wlmData.dll.GetPatternDataNum.return_value = -6  # not available

        with HighFinesse_Wlm(self.ctx, "wlm") as wlm:
            with self.assertRaises(QMI_InstrumentException):
                wlm.get_data_pattern_data(1, 1, 2048)



""" QMI instrument driver for the PicoQuant MultiHarp 150 instrument.

The instrument driver makes use of the manufacturer provided software libraries, "mhlib.so" for Linux OS,
or "mhlib.dll" or "mhlib64.dll" for 32-bit and 64-bit Windows OS, respectively.
Please find the licence terms for these files in the dedicated software package for the MultiHarp instrument at
https://www.picoquant.com/products/category/tcspc-and-time-tagging-modules/multiharp-150-high-throughput-multichannel-event-timer-tcspc-unit
--> "Software" tab --> download link in "Current software and developer's library version".
"""
import ctypes
import enum
import logging
from fractions import Fraction
from typing import List, Tuple

import numpy as np

from qmi.core.exceptions import QMI_InvalidOperationException
from qmi.core.rpc import rpc_method
from qmi.instruments.picoquant.support._library_wrapper import _LibWrapper
from qmi.instruments.picoquant._picoquant import _str_to_enum, _PicoquantHarp, _EDGE
from qmi.instruments.picoquant.support._events import _MODE

_logger = logging.getLogger(__name__)


@enum.unique
class _REFSRC(enum.Enum):
    """Symbolic constants for the :func:`~MultiHarpDevice.initialize` method's `refsource` argument.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """
    INTERNAL = 0
    """Use internal clock."""
    EXTERNAL_10MHZ = 1
    """Use 10 MHz external clock."""
    WR_MASTER_GENERIC = 2
    """White Rabbit master with generic partner."""
    WR_SLAVE_GENERIC = 3
    """White Rabbit slave with generic partner."""
    WR_GRANDM_GENERIC = 4
    """White Rabbit grand master with generic partner."""
    EXTN_GPS_PPS = 5
    """Use 10 MHz + PPS from GPS."""
    EXTN_GPS_PPS_UART = 6
    """Use 10 MHz + PPS + time via UART from GPS."""
    WR_MASTER_MHARP = 7
    """White Rabbit master with MultiHarp as partner."""
    WR_SLAVE_MHARP = 8
    """White Rabbit slave with MultiHarp as partner."""
    WR_GRANDM_MHARP = 9
    """White Rabbit grand master with MultiHarp as partner."""


@enum.unique
class _MEASCTL(enum.Enum):
    """Symbolic constants for the :func:`~MultiHarpDevice.setMeasurementControl` `control` argument.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """
    SINGLESHOT_CTC = 0
    """Default value.
       Acquisition starts by software command and runs until CTC expires. 
       The duration is set by the ``tacq`` parameter passed to ``MH_StartMeas``.
    """
    C1_GATED = 1
    """Data is collected for the period where C1 is active. 
       This can be the logical high or low period dependent
       on the value supplied to the parameter ``startedge``.
    """
    C1_START_CTC_STOP = 2
    """Data collection is started by a transition on C1 and
       stopped by expiration of the internal CTC. 
       Which transition actually triggers the start is given
       by the value supplied to the parameter ``startedge``. 
       Theduration is set by the ``tacq``parameter passed to MH_StartMeas.
    """
    C1_START_C2_STOP = 3
    """Data collection is started by a transition on C1 and stopped
       by a transition on C2. Which transitions actually trigger
       start and stop is given by the values supplied to the parameters startedge and ``stopedge```
    """
    WR_M2S = 4
    """Sets measurement conrol for a pair of devices such that the master device
       not only starts its own measurement but also that of the slave.
       This setting must be applied identivcally for both devices and it requires
       that the two devices have established a white rabbit connection.
    """
    WR_S2M = 5
    """Sets measurement conrol for a pair of devices such that the master device
       not only starts its own measurement but also that of the slave.
       This setting must be applied identivcally for both devices and it requires
       that the two devices have established a white rabbit connection. 
    """
    SW_START_SW_STOP = 6  # new since v3.1
    """Sets measurement control such that the duration of a measurement is purely under software control,
       i.e. started by te call of ``MH_StartMeas`` and stopped by the call of ``MH_StartMeas``. 
       This overcomes the time limits of the hardware CTC and allows virtually unlimited measurement times.
       The resulting shortcoming is that the resulting duration is then less precise.
    """

@enum.unique
class _FLAG(enum.Enum):
    """Bitfield constants for the return value of the :func:`~MultiHarpDevice.getFlags` function.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """
    OVERFLOW = 0x0001
    """Histogram mode only. It indicates that a histogram measurement has reached the maximum count 
    as specified via MH_SetStopOverflow."""
    FIFOFULL = 0x0002
    """TTTR mode only. It indicates that the data FiFo has run full. The measurement will then have
    to be aborted as data integrity is no longer maintained."""
    SYNC_LOST = 0x0004
    """This flag may occur in T3 mode and in histo mode. It indicates that the sync signal has been lost
    which in this case is critical as the function of T3 mode and histo mode relies on an uninterrupted
    sync signal."""
    REF_LOST = 0x0008
    """This flag will occur when the MultiHarp is programmed to use an external reference clock and this
    reference clock is lost."""
    SYSERROR = 0x0010
    """Indicates an error of the hardware or internal software. The user should in this case
    call the library routine MH_GetDebugInfo and provide the result to PicoQuant support."""
    ACTIVE = 0x0020
    """Measurement is running."""
    CNTS_DROPPED = 0x0040
    """Indicates that counts were dropped at the first level FiFo directly following the TDC of an input channel.
    This occurs typically only at extremely high count rates. Dependent on the application scenario this may or
    may not be considered critical."""


@enum.unique
class _WARNING(enum.Enum):
    """Bitfield constants for the return value of the :func:`~MultiHarpDevice.getWarnings` function.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file. For full descriptions
    see the chapter '8.1. Warnings' from the MultiHarp MHLib Manual.
    """
    SYNC_RATE_ZERO = 0x0001
    """No pulses are detected at the sync input. In histogramming and T3 mode this is crucial and the
    measurement will not work without this signal."""
    SYNC_RATE_VERY_LOW = 0x0002
    """The detected pulse rate at the sync input is below 100 Hz and cannot be determined accurately.
    Other warnings may not be reliable under this condition."""
    SYNC_RATE_TOO_HIGH = 0x0004
    """The pulse rate at the sync input (after the divider) is higher than 75 MHz. This is close to 
    the TDC limit. Sync events will be lost above 78 MHz."""
    INPT_RATE_ZERO = 0x0010
    """No counts are detected at any of the input channels."""
    INPT_RATE_TOO_HIGH = 0x0040
    """The overall pulse rate at the input channels is higher than 80 MHz (USB 3.0 connection) or higher
    than 9 MHz (USB 2.0 connection). The measurement will likely lead to a FIFO overrun."""
    INPT_RATE_RATIO = 0x0100
    """This warning is issued in histogramming and T3 mode when the rate at any input channel is higher
    than 5% of the sync rate."""
    DIVIDER_GREATER_ONE = 0x0200
    """In T2 mode: The sync divider is set larger than 1. This is probably not intended.
    In histogramming and T3 mode: If the pulse rate at the sync input is below 75 MHz then a divider
    >1 is not needed."""
    TIME_SPAN_TOO_SMALL = 0x0400
    """This warning is issued in histogramming and T3 mode when the sync period (1/SyncRate) is longer
    than the start to stop time span that can be covered by the histogram or by the T3 mode records."""
    OFFSET_UNNECESSARY = 0x0800
    """This warning is issued in histogramming and T3 mode when an offset >0 is set even though the sync
    period (1/SyncRate) can be covered by the measurement time span without using an offset."""
    DIVIDER_TOO_SMALL = 0x1000
    """The pulse rate at the sync input (after the divider) is higher than 75 MHz."""
    COUNTS_DROPPED = 0x2000
    """This warning is issued when the front end of the data processing pipeline was not able to process
    all events that came in."""


class _WR_STATUS(enum.Enum):
    """ White Rabbit subsystem status bit-masks.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """

    # status_link_on       1 bit  : bit  0
    # status_link_up       1 bit  : bit  1
    # status_mode          2 bits : bits 3..2
    # status_locked_calibd 1 bit  : bit  4
    # status_ptp           3 bits : bits 7..5
    # status_servo         3 bits : bits 10..8
    # status_mac_set       1 bit  : bit  11
    # (reserved)          19 bits : bits 30..12
    # status_is_new        1 bit  : bit  31

    LINK_ON = 0x00000001
    """WR link is switched on."""
    LINK_UP = 0x00000002
    """WR link is established."""

    MODE_BITMASK = 0x0000000C
    """Mask for the mode bits."""
    MODE_OFF = 0x00000000
    """Mode is "off"."""
    MODE_SLAVE = 0x00000004
    """Mode is "slave"."""
    MODE_MASTER = 0x00000008
    """Mode is "master"."""
    MODE_GMASTER = 0x0000000C
    """Mode is "grandmaster"."""

    LOCKED_CALIBD = 0x00000010
    """Locked and calibrated."""

    PTP_BITMASK = 0x000000E0
    """Mask for the PTP bits."""
    PTP_LISTENING = 0x00000020
    """(Undocumented)"""
    PTP_UNCLWRSLCK = 0x00000040
    """(Undocumented)"""
    PTP_SLAVE = 0x00000060
    """(Undocumented)"""
    PTP_MSTRWRMLCK = 0x00000080
    """(Undocumented)"""
    PTP_MASTER = 0x000000A0
    """(Undocumented)"""

    SERVO_BITMASK = 0x00000700
    """Mask for the servo bits."""
    SERVO_UNINITLZD = 0x00000100
    """(Undocumented)"""
    SERVO_SYNC_SEC = 0x00000200
    """(Undocumented)"""
    SERVO_SYNC_NSEC = 0x00000300
    """(Undocumented)"""
    SERVO_SYNC_PHASE = 0x00000400
    """(Undocumented)"""
    SERVO_WAIT_OFFST = 0x00000500
    """(Undocumented)"""
    SERVO_TRCK_PHASE = 0x00000600
    """(Undocumented)"""

    MAC_SET = 0x00000800
    """User-defined MAC address is set."""
    IS_NEW = 0x80000000
    """Status updated since last check."""


class PicoQuant_MultiHarp150(_PicoquantHarp):
    """Instrument driver for the PicoQuant MultiHarp 150."""

    _MODEL = "MH"
    _MAXDEVNUM = 8
    _TTREADMAX = 1048576

    @property
    def _max_dev_num(self):
        return self._MAXDEVNUM

    @property
    def _ttreadmax(self):
        return self._TTREADMAX

    @property
    def _model(self):
        return self._MODEL

    @property
    def _lib(self) -> _LibWrapper:
        if self._lazy_lib is None:  # type: ignore
            self._lazy_lib = _LibWrapper('MH')
        return self._lazy_lib

    @rpc_method
    def initialize(self, mode_str: str, refsource_str: str) -> None:
        """Initialize the device.

        This routine must be called before any of the other methods, except OpenDevice, CloseDevice,
        GetErrorString and GetLibraryVersion can be used.

        See the manual for more information on measurement modes, external clock, and
        White Rabbit (WR) functionality.

        Selecting WR as a clock source requires that a WR connection has actually been established beforehand.
        Unless the WR connection is established by a WR startup script this will require a two stage process
        initially initializing with internal clock source, then setting up the WR connection by means of the WR
        routines described below, then initializing again with the desired WR clock model.

        Arguments:
            mode_str (str): Opening mode. Can be any of 'HIST', 'T2', 'T3'.
                            The latest driver version V3.0+ supports T2 and T3 modes.
            refsource_str (str): Reference source for time. Can be any of 'INTERNAL', 'EXTERNAL_10MHZ',
                'WR_MASTER_GENERIC', 'WR_SLAVE_GENERIC', 'WR_GRANDM_GENERIC', 'EXTN_GPS_PPS', 'EXTN_GPS_PPS_UART',
                'WR_MASTER_MHARP', 'WR_SLAVE_MHARP', 'WR_GRANDM_MHARP'.

        Raises:
            ValueError: if either argument is an unknown string value.
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()

        if self._measurement_running:
            raise QMI_InvalidOperationException("Measurement still active")
        mode = _str_to_enum(_MODE, mode_str)
        if mode not in (_MODE.HIST, _MODE.T2, _MODE.T3):
            raise ValueError("Unsupported measurement mode")

        if self._lib_version.startswith("2") and mode is _MODE.T3:
            raise ValueError("Unsupported measurement mode")

        refsource = _str_to_enum(_REFSRC, refsource_str)

        with self._device_lock:
            self._lib.Initialize(self._devidx, mode.value, refsource.value)
            self._mode = mode

        # Reset the event filter configuration.
        self.set_event_filter(reset_filter=True)
        self.set_block_events(False)
        _logger.info("[%s] Initialized device with mode %s and source %s", self._name, mode.value, refsource.value)

    @rpc_method
    def get_module_info(self) -> List[Tuple[int, int]]:
        self._check_is_open()
        info = []
        with self._device_lock:
            nummod = ctypes.c_int()
            self._lib.GetNumOfModules(self._devidx, nummod)
            for modidx in range(nummod.value):
                modelcode = ctypes.c_int()
                versioncode = ctypes.c_int()
                self._lib.GetModuleInfo(self._devidx, modidx, modelcode, versioncode)
                info.append((modelcode.value, versioncode.value))
        return info

    @rpc_method
    def set_sync_channel_offset(self, value: int) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.SetSyncChannelOffset(self._devidx, value)

    @rpc_method
    def set_sync_edge_trigger(self, level: int, edge_str: str) -> None:
        """Set SYNC trigger level and edge.

        Arguments:
            level: Trigger level, in mV. A value ranging from :-1200 mV to +1200 mV.
            edge_str: Either 'RISING' or 'FALLING'.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        edge = _str_to_enum(_EDGE, edge_str)
        with self._device_lock:
            self._lib.SetSyncEdgeTrg(self._devidx, level, edge.value)

    @rpc_method
    def set_sync_dead_time(self, on: int, deadtime: int) -> None:
        """Set sync channel dead-time.

        Arguments:
            on: 0 = set minimal dead-time, 1 = activate extended dead-time.
            deadtime: extended dead-time in ps. Valid range from `constants.EXTDEADMIN` (800 ps) to
                `constants.EXTDEADMAX` (160000 ps).

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetSyncDeadTime(self._devidx, on, deadtime)

    @rpc_method
    def set_input_edge_trigger(self, channel: int, level: int, edge_str: str) -> None:
        """Set input channel trigger level and edge.

        Arguments:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            level: Trigger level, in mV. A value ranging from -1200 mV to +1200 mV.
            edge_str: 'RISING' or 'FALLING'.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            edge = _str_to_enum(_EDGE, edge_str)
            self._lib.SetInputEdgeTrg(self._devidx, channel, level, edge.value)

    @rpc_method
    def set_input_channel_dead_time(self, channel: int, on: int, deadtime: int) -> None:
        """Set input channel dead-time.

        Arguments:
            channel: Channel index (range from 0 to `number_of_channels` - 1).
            on: 0 = set minimal dead-time, 1 = activate extended dead-time.
            deadtime: extended dead-time in ps. Valid range from `constants.EXTDEADMIN` (800 ps) to
                `constants.EXTDEADMAX` (160000 ps).

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            self._lib.SetInputDeadTime(self._devidx, channel, on, deadtime)

    @rpc_method
    def set_measurement_control(self, meascontrol_str: str, startedge_str: str, stopedge_str: str) -> None:
        self._check_is_open()
        meascontrol = _str_to_enum(_MEASCTL, meascontrol_str)
        startedge = _str_to_enum(_EDGE, startedge_str)
        stopedge = _str_to_enum(_EDGE, stopedge_str)
        with self._device_lock:
            self._lib.SetMeasControl(self._devidx, meascontrol.value, startedge.value, stopedge.value)

    @rpc_method
    def set_trigger_output(self, period: int) -> None:
        self._check_is_open()
        with self._device_lock:
            self._lib.SetTriggerOutput(self._devidx, period)

    @rpc_method
    def get_flags(self) -> List[str]:
        self._check_is_open()
        with self._device_lock:
            bitset = ctypes.c_int()
            self._lib.GetFlags(self._devidx, bitset)

            return [flag.name for flag in _FLAG if bitset.value & flag.value]

    @rpc_method
    def get_start_time(self) -> Fraction:
        """Get start time of measurement.

        Retrieve the start time of a measurement with picosecond resolution. It relates always to the start of the
        most recent measurement, be it completed or only just started. The result is to be interpreted in the sense
        of a unix time, i.e. elapsed picoseconds since January 1st 1970 00:00:00 UTC (Universal Time).

        Note that the actual resolution is the device's base resolution. Actual accuracy depends on the chosen time
        base, e.g., a White Rabbit grandmaster can be very accurate.

        With less accurate clocks the high resolution result can still be meaningful in a relative sense,
        e.g. between two devices synchronized over White Rabbit. With internal clocking the accuracy only reflects
        that of the PC clock.

        Returns:
            Start time as a Fraction, to ensure that no precision is lost.

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        self._check_is_open()
        with self._device_lock:
            timedw2 = ctypes.c_uint()
            timedw1 = ctypes.c_uint()
            timedw0 = ctypes.c_uint()
            self._lib.GetStartTime(self._devidx, timedw2, timedw1, timedw0)

        timedw2_int = timedw2.value
        timedw1_int = timedw1.value
        timedw0_int = timedw0.value
        picoseconds = (timedw2_int << 64) | (timedw1_int << 32) | timedw0_int
        starttime = Fraction(picoseconds, 1000000000000)
        _logger.debug("[%s] Measurement start-time fraction is %s", self._name, starttime)
        return starttime

    @rpc_method
    def get_warnings(self) -> List[str]:
        self._check_is_open()
        with self._device_lock:
            bitset = ctypes.c_int()
            self._lib.GetWarnings(self._devidx, bitset)
            return [warning.name for warning in _WARNING if bitset.value & warning.value]

    @rpc_method
    def get_all_count_rates(self) -> Tuple[int, Tuple[int, ...]]:
        """Get count rates of SYNC and INPUT channels.

        This can be used as a replacement of :func:`get_sync_rate()` and :func:`get_count_rate()`
        when all rates need to be retrieved in an efficient manner.

        Returns:
            A tuple (syncrate, inputrates), where syncrate is the sync channel rate (integer),
            and inputrate is a tuple of integers (one per channel).

        Raises:
            QMI_InstrumentException: in case of a library error.
        """
        with self._device_lock:
            nchannels = ctypes.c_int()
            self._lib.GetNumOfInputChannels(self._devidx, nchannels)
            num_channels = nchannels.value
            syncrate = ctypes.c_int()
            inputrates = (ctypes.c_int * num_channels)()
            self._lib.GetAllCountRates(self._devidx, syncrate, inputrates)
            syncrate_int = syncrate.value
            inputrates_tuple = tuple((int(i) for i in inputrates))
            return syncrate_int, inputrates_tuple

    @rpc_method
    def get_all_histograms(self) -> np.ndarray:
        self._check_is_open()
        channels = self.get_number_of_input_channels()
        with self._device_lock:
            all_histdata = np.empty((channels, self._actuallen), dtype=np.uint32)
            ctypes_all_histdata = all_histdata.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32))
            self._lib.GetAllHistograms(self._devidx, ctypes_all_histdata)
            return all_histdata.copy()



"""Instrument driver for the NewPort/NewFocus TLB-670x tunable laser controller.

This driver depends on the usbdll.dll library from New Focus. The permission for linking to the DLL library file, with
a proprietary license, has been explicitly granted by Newport Corporation to QuTech.
"""
import ctypes
import logging
import re
import sys
import time
from typing import Any, List, Tuple, Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method

from .tlb670x_error_messages import ERROR_MESSAGES


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class NewFocus_TLB670X(QMI_Instrument):
    """Instrument driver for the New Focus TLB-670x USB tunable laser controller (Windows platforms only.)

    This driver requires that the DLL provided by NewFocus ("usbdll.dll") is on the DLL search path.

    This class contains a number of functions. Most importantly, the _send and _receive functions allow for sending
    strings and receiving the TLC response data via a buffer, respectively. The various command strings
    are invoked through functions defined in this class. The most important functions are the get_- and
    set_wavelength that can be used to change the TLC setpoint.
    """

    # Shared library file name.
    DLL_NAME = "usbdll.dll"

    # USB speed (baud).
    USB_SPEED = 480e6

    # Sleep time after reset (seconds).
    RESET_SLEEP_TIME = 7

    # Maximum buffer length (bytes).
    MAX_BUFFER_LENGTH = 64

    # Device identification.
    VENDOR_ID = 0x104D
    PRODUCT_ID = 0x100A

    def __init__(self, context: QMI_Context, name: str, serial_number: Union[int, str]) -> None:
        """Initialize the instrument driver.

        Parameters:
            context:        QMI_Context for the instrument
            name:           Name for this instrument instance.
            serial_number:  The controller's serial number in format <SNabcd>, where a, b, c, d are digits 0-9. Not
                            including 'SN' or providing the serial number as integer works too, but is discouraged.
        """
        # Check platform first.
        if not sys.platform.startswith("win"):
            raise RuntimeError("The NewFocus TLB670x QMI driver is only supported on Windows platforms!")

        # Initialize driver superclass.
        super().__init__(context, name)

        # Prepare handle to DLL and device; set in open().
        self._handle: Any = ctypes.WinDLL(self.DLL_NAME)
        self._device_id: int = -1  # Initialize with non-existing device number. The real values are in range [0-31].

        # Generously parse serial number.
        serial_number = str(serial_number)
        if not serial_number.startswith("SN"):
            _logger.warning("Serial number should start with SN, but got %s", serial_number)
            serial_number = "SN" + str(serial_number)
        self.serial_number: str = serial_number

    def _send(self, command_string: str) -> str:
        """Sends the given command to the device and returns the response as a tuple (response_code, response_value),
        where a non-zero `response_code` indicates that the device reported that executing the command failed. In this
        case, `response_value` will hold the error message. If the command was executed successfully, `response_code`
        will be 0 and `response_value` will be the response to the command as a string.

        Parameters:
            command_string: Command string to be parsed for sending.
        Returns:
            The controller response.
        Raises:
            QMI_InstrumentException if the controller returned an error code.
        """
        self._check_is_open()
        _logger.debug("Sending message to device: %s", command_string)

        # Prepare command.
        command = ctypes.create_string_buffer(self.MAX_BUFFER_LENGTH)
        command.value = command_string.encode()
        command_length = ctypes.c_ulong(len(command.value))

        # Send command and check for any errors.
        result = self._handle.newp_usb_send_ascii(self._device_id, ctypes.byref(command), command_length)
        if result != 0:
            # Command failed.
            error_msg = ERROR_MESSAGES.get(result, "Unknown error")
            _logger.error("Received error response: %s (%d)", error_msg, result)
            self._reinit_device()  # needed to clear error state
            raise QMI_InstrumentException(f"Command {command_string} returned an error {result}: {error_msg}")

        # Query for response.
        time.sleep(1.0 / self.USB_SPEED * command_length.value)
        response = self._receive()
        if "?" in command_string:
            # Check if response string consists of multiple responses, separated by "\r\n".
            if len(response) > 1 and response[-1] != "OK":
                # The response probably is this. But check for sporadic "*IDN?" query response first.
                idn_match = re.search(r"v\d.\d(.\d)? [\d/]+ SN[\d]+", response[-1])
                if idn_match:  # The entry is the invalid response, delete it
                    del response[-1]

                # OK check. Otherwise checks response[0].
                if response[-1] != "OK" and response[-1] != "+00":
                    return response[-1]

            if response[0] == "OK":
                # Otherwise, we did not get the correct response to our query, try to re-read once
                time.sleep(1.0 / self.USB_SPEED * command_length.value)
                response = self._receive()

        return response[0]

    def _receive(self) -> List[str]:
        """Query the controller for a command response.

        Returns:
            List of response strings from the controller.
        Raises:
            QMI_InstrumentException if the device returns an error code.
        """
        # Prepare buffer.
        buf_length = ctypes.c_ulong(self.MAX_BUFFER_LENGTH)
        buf = ctypes.create_string_buffer(buf_length.value)
        bytes_read = ctypes.c_ulong(64)

        # Query controller and check for any errors.
        result = self._handle.newp_usb_get_ascii(
            self._device_id, ctypes.byref(buf), buf_length, ctypes.byref(bytes_read)
        )
        if result != 0:
            error = ERROR_MESSAGES.get(result, "Unknown error")
            _logger.error("Received error response: %s (%d)", error, result)
            self._reinit_device()  # needed to clear error state
            raise QMI_InstrumentException(f"No response from device: {error} ({result})")

        # Decode response.
        response = buf.value.decode().split("\r\n")
        _logger.debug("Response from device: %s", response)
        # Pop out empty strings
        response = [r for r in response if r != ""]
        return response

    def _get_device_info(self) -> List[Tuple[int, str]]:
        """Get the contents of the device information buffer (a list of deviceID,deviceDescription pairs).

        Raises:
            QMI_InstrumentException: In cases where getting the device info fails for some reason.

        Returns:
            device_info: The found devices' info as a list of tuples (device ID#, device description string).
        """
        # Prepare buffer.
        buf_length = ctypes.c_ulong(self.MAX_BUFFER_LENGTH)
        buf = ctypes.create_string_buffer(buf_length.value)

        # Query controller and check for any errors.
        result = self._handle.newp_usb_get_device_info(ctypes.byref(buf))
        if result != 0:
            # Cannot do reinit here, because this method is used in init_device(); only raise exception that needs to
            # be handled by the caller.
            error = ERROR_MESSAGES.get(result, "Unknown error")
            raise QMI_InstrumentException(f"Unable to load device info: {error}")

        # Decode response; this is a ;-separated list of <deviceID>,<deviceDescription> pairs.
        response: str = buf.value.decode()
        device_info = []
        if response:
            for device in response.split(';'):
                if ',' in device:
                    device_id_str, device_descr_string = device.split(',')
                    device_info.append(
                        (int(device_id_str), device_descr_string)
                    )

        return device_info

    def _init_device(self) -> None:
        """Find and initialize the device ID for the device with the given serial number. For re-initialization no
        SN number check is done.

        Attributes:
            self._device_id: Set the device ID attribute if an instrument with matching serial number is found.
        Raises:
            QMI_InstrumentException: If no device with the correct serial number is found.
        """
        # (Re-)Open all USB devices that match the USB product ID.
        self._handle.newp_usb_init_product(self.PRODUCT_ID)

        # Get list of all devices found.
        device_info = self._get_device_info()
        if not device_info:
            self._uninit_device()
            raise QMI_InstrumentException("No TLB-670X instrument present")

        # Check if a device has already been found previously. If yes, no need to re-check with serial number.
        if self._device_id in range(32):
            return

        # The deviceDescription contains the serial number; find the one that matches our serial number.
        _logger.debug("Devices found: %s", device_info)
        for device_id, device_descr in device_info:
            serial_number_match = re.search(r"SN\d+", device_descr)
            if serial_number_match is not None and serial_number_match[0] == self.serial_number:
                self._device_id = device_id
                _logger.debug("Match: device_id=%d", device_id)
                return

        # Device not found.
        self._uninit_device()
        raise QMI_InstrumentException(f"No device with serial_number={self.serial_number} was found")

    def _uninit_device(self) -> None:
        """Uninitialize the device."""
        self._handle.newp_usb_uninit_system()

    def _reinit_device(self) -> None:
        """Reinitialize the device, clearing any errors."""
        self._check_is_open()
        _logger.debug("Reinitializing device to clear error state...")
        self._uninit_device()
        self._init_device()

    @rpc_method
    def open(self) -> None:
        """Open connection to the device controller."""
        super().open()
        _logger.info(f"Opening connection to {self._name}")
        self._init_device()

    @rpc_method
    def close(self) -> None:
        """Close connection to the device controller."""
        super().close()
        _logger.info(f"Closing connection to {self._name}")
        self._uninit_device()

    @rpc_method
    def get_available_devices_info(self) -> List[Tuple[int, str]]:
        """Get a list of all available devices that match the product ID for the TLB670-X."""
        self._check_is_open()
        try:
            device_info = self._get_device_info()
        except QMI_InstrumentException as exc:
            self._reinit_device()  # needed to clear error state
            raise exc

        return device_info

    @rpc_method
    def get_ident(self) -> QMI_InstrumentIdentification:
        """Get device identification."""
        return QMI_InstrumentIdentification(
            vendor="NewFocus", model="TLB670X", serial=self.serial_number, version=None
        )

    @rpc_method
    def get_device_id(self) -> int:
        """Get the device ID of the device."""
        self._check_is_open()
        return self._device_id

    @rpc_method
    def reset(self) -> None:
        """Reset the device."""
        _logger.info(f"Resetting {self._name}")

        # First check for errors.
        self.check_error_status()

        # Send reset command.
        try:
            self._send("*RST")
        except QMI_InstrumentException as exc:
            raise exc

        time.sleep(self.RESET_SLEEP_TIME)

    @rpc_method
    def check_error_status(self) -> str:
        """Check the error status of the device and return the error string if any.

        Returns:
            The error query response.
        """
        response = self._send("ERRSTR?")
        _logger.info(f"Error query response: {response}")
        return response

    @rpc_method
    def get_wavelength(self) -> float:
        """Get the current wavelength setpoint.

        Returns:
            The wavelength setpoint in nanometers.
        """
        response = self._send("SOURce:WAVElength?")
        return float(response)

    @rpc_method
    def set_wavelength(self, wavelength: float) -> None:
        """Set wavelength.

        Parameters:
            wavelength: wavelength in nanometers.
        """
        if wavelength <= 0:
            raise ValueError(f"Wavelength must be a positive number; got {wavelength}")

        self._send(f"SOURce:WAVElength {wavelength}")

    @rpc_method
    def get_powermode(self) -> int:
        """Get power state.

        Returns:
            Power output state as 0: laser off; 1: laser on.
        """
        response = self._send("OUTPut:STATe?")
        return int(response)

    @rpc_method
    def set_powermode(self, powermode: int) -> None:
        """Set power state.

        Parameters:
            powermode:  0 turns the laser off, 1 turns the laser on.
        """
        if powermode not in (0, 1):
            raise ValueError(f"Power mode must be 0 or 1; got {powermode}")

        self._send(f"OUTPut:STATe {powermode}")

    @rpc_method
    def get_trackingmode(self) -> int:
        """Get tracking state.

        Returns:
            Tracking state as 0: disabled; 1: enabled.
        """
        response = self._send("OUTPut:TRACk?")
        return int(response)

    @rpc_method
    def set_trackingmode(self, trackingmode: int) -> None:
        """Set tracking state.

        Parameters:
            trackingmode:  0: disable, 1: enable.
        """
        if trackingmode not in (0, 1):
            raise ValueError(f"Tracking mode must be 0 or 1; got {trackingmode}")

        self._send(f"OUTPut:TRACk {trackingmode}")

    @rpc_method
    def get_diode_current(self) -> float:
        """Get the laser diode current.

        Returns:
            Laser diode current in mA.
        """
        response = self._send("SENSe:CURRent:DIODe")
        return float(response)

    @rpc_method
    def set_diode_current(self, current: Union[int, float, str]) -> None:
        """Set the laser diode current in mA.

        Parameters:
            current:    either a float representing the desired current in mA, or the string literal 'MAX' to set the
                        current to the maximum allowable value.
        """
        if isinstance(current, (int, float)):
            if current <= 0:
                raise ValueError(f"Diode current must be a positive value; got {current}")
        elif isinstance(current, str):
            current = current.upper()
            if current != "MAX":
                raise ValueError(f"Invalid string for diode current setpoint; must be 'MAX', got {current}")
        else:
            raise ValueError(f"Invalid value for diode current setpoint: {current}")

        self._send(f"SOURce:CURRent:DIODe {current}")

    @rpc_method
    def get_piezo_voltage(self) -> float:
        """Retrieve the current piezo voltage.

        Returns:
            Piezo voltage as percentage of its range.
        """
        response = self._send("SOURce:VOLTage:PIEZo?")
        return float(response)

    @rpc_method
    def set_piezo_voltage(self, percentage: float) -> None:
        """Set the piezo voltage as percentage of its range.

        Parameters:
            percentage: piezo voltage percentage.
        """
        if not 0 <= percentage <= 100.0:
            raise ValueError(f"Percentage should be in range 0..100; got {percentage}")

        self._send(f"SOURce:VOLTage:PIEZo {percentage}")



"""Unit-tests for WL Photonics WLTF-N driver class."""
import logging
import unittest
from unittest.mock import Mock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_SerialTransport
from qmi.core.usbtmc import Instrument

from qmi.instruments.wl_photonics import WlPhotonics_WltfN


class TestWlPhotonicsWltfNOpenCloseSerial(unittest.TestCase):

    def setUp(self):
        logging.getLogger("qmi.instruments.wl_photonics.wltf_n").setLevel(logging.CRITICAL)
        qmi.start("TestWlPhotonicsWltfNClassContext")
        # Add patches
        patcher = patch('qmi.instruments.wl_photonics.wltf_n.create_transport', spec=QMI_SerialTransport)
        self._transport_mock = patcher.start()
        self.addCleanup(patcher.stop)
        # Make DUT
        self.instr: WlPhotonics_WltfN = WlPhotonics_WltfN(qmi.context(), "WlPhotonics_wltf_n", "")

    def tearDown(self):
        qmi.stop()
        logging.getLogger("qmi.instruments.wl_photonics.wltf_n").setLevel(logging.NOTSET)

    def test_open_close(self):
        """open() command not just opens the instrument, but immediately also checks the wavelength/steps
        range and the power unit to set the power level range. Test the whole shebang.

        close() command then just closes the connection to the instrument and the transport.
        """
        # Initially, the wavelength and power range should be bare data classes
        expected_initial_out = "<class 'qmi.instruments.wl_photonics.wltf_n._{rng}Range'>"
        self.assertEqual(expected_initial_out.format(rng="Wavelength"), str(self.instr._wavelength_range))
        self.assertEqual(expected_initial_out.format(rng="Steps"), str(self.instr._steps_range))
        # Arrange
        expected_wl_min = 1021.509
        expected_wl_max = 1072.505
        step_min = 556
        step_max = 4654
        expected_start_step = (step_max - step_min) // 2

        self.instr._transport.read_until.side_effect = [
            # output of dev?
            b'WL200: SN(201307374), MD(2018-11-23)\r\n' +
            b'WL Range: 1021.509~1072.505nm(Step: 4654~556)\r\n' +
            b'OK\r\n',
            # output of s?
            f"Step: {expected_start_step}\r\n, ERR: -1, Status: 0x340880".encode() +
            b'OK\r\n',
        ]
        expected_calls = [
            call().open(),
            call().write(b"dev?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT),
            call().write(b"s?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT)
        ]
        # Act
        self.instr.open()
        # Assert
        self._transport_mock.assert_called_once()
        self.assertEqual(expected_wl_min, self.instr._wavelength_range.min)
        self.assertEqual(expected_wl_max, self.instr._wavelength_range.max)
        self.assertEqual(step_min, self.instr._steps_range.min)
        self.assertEqual(step_max, self.instr._steps_range.max)
        self.assertEqual(expected_start_step, self.instr._motor_position)
        self._transport_mock.assert_has_calls(expected_calls, any_order=True)
        # Test close
        self.instr.close()
        self._transport_mock.assert_called_once()


class TestWlPhotonicsWltfNOpenClosePyUSB(unittest.TestCase):

    def setUp(self):
        logging.getLogger("qmi.instruments.wl_photonics.wltf_n").setLevel(logging.CRITICAL)
        qmi.start("TestWlPhotonicsWltfNClassContext")
        # Add patches - now we should patch a bit deeper to test the creation of USBTMC transport without SN.
        patcher = patch('qmi.core.usbtmc.Instrument', autospec=Instrument)
        self._transport_mock = patcher.start()
        self.addCleanup(patcher.stop)
        # Make DUT
        self.vendor_id = 0x10C4
        self.product_id = 0xEA60
        serial = ""  # No serial number known. Does it still work?
        transport_string = f"usbtmc:vendorid={self.vendor_id}:productid={self.product_id}:serialnr={serial}"
        with patch("sys.platform", "linux1"):
            self.instr: WlPhotonics_WltfN = WlPhotonics_WltfN(qmi.context(), "WlPhotonics_wltf_n", transport_string)

    def tearDown(self):
        qmi.stop()
        logging.getLogger("qmi.instruments.wl_photonics.wltf_n").setLevel(logging.NOTSET)

    def test_open_close(self):
        """open() command not just opens the instrument, but immediately also checks the wavelength/steps
        range and the power unit to set the power level range. Test the whole shebang.

        close() command then just closes the connection to the instrument and the transport.
        """
        # Initially, the wavelength and power range should be bare data classes
        expected_initial_out = "<class 'qmi.instruments.wl_photonics.wltf_n._{rng}Range'>"
        self.assertEqual(expected_initial_out.format(rng="Wavelength"), str(self.instr._wavelength_range))
        self.assertEqual(expected_initial_out.format(rng="Steps"), str(self.instr._steps_range))
        # Arrange
        expected_wl_min = 1021.509
        expected_wl_max = 1072.505
        step_min = 556
        step_max = 4654
        expected_start_step = (step_max - step_min) // 2

        self.instr._transport._device = self._transport_mock
        self._transport_mock().read_raw.side_effect = [
            # output of dev?
            b'WL200: SN(201307374), MD(2018-11-23)\r\n' +
            b'WL Range: 1021.509~1072.505nm(Step: 4654~556)\r\n' +
            b'OK\r\n',
            # output of s?
            f"Step: {expected_start_step}\r\n, ERR: -1, Status: 0x340880".encode() +
            b'OK\r\n',
        ]
        expected_calls = [
            call(self.vendor_id, self.product_id, ''),  # QMI_PyUsbTransport __init__()
            call().open(),  # QMI_PyUsbTransport.open()
            call().write_raw(b"dev?\r\n"),  # dev? call in open()
            call().write_raw(b"s?\r\n")  # s? call in open()
        ]
        # Act
        self.instr.open()
        # Assert
        self.assertEqual(2, self._transport_mock.call_count)  # self._transport_mock()... is one call, __init__() 2nd
        self.assertEqual(expected_wl_min, self.instr._wavelength_range.min)
        self.assertEqual(expected_wl_max, self.instr._wavelength_range.max)
        self.assertEqual(step_min, self.instr._steps_range.min)
        self.assertEqual(step_max, self.instr._steps_range.max)
        self.assertEqual(expected_start_step, self.instr._motor_position)
        self._transport_mock.assert_has_calls(expected_calls, any_order=True)
        # Test close
        self.instr.close()
        self.assertEqual(2, self._transport_mock.call_count)


class TestWlPhotonicsWltfNClassMethods(unittest.TestCase):

    def setUp(self):
        # Make as standard to open in wavelength and dBm mode.
        self.wl_min = 1021.509
        self.wl_max = 1072.505
        self.step_min = 556
        self.step_max = 4654
        expected_start_step = (self.step_max - self.step_min) // 2

        logging.getLogger("qmi.instruments.wl_photonics.wltf_n").setLevel(logging.CRITICAL)
        qmi.start("TestWlPhotonicsWltfNClassContext")
        # Add patches
        patcher = patch('qmi.instruments.wl_photonics.wltf_n.create_transport', spec=QMI_SerialTransport)
        self._transport_mock = patcher.start()
        self.addCleanup(patcher.stop)
        self._transport_mock.write = Mock()
        # Make DUT
        self.instr: WlPhotonics_WltfN = WlPhotonics_WltfN(qmi.context(), "RSB100a", "")
        self.instr._transport.read_until.side_effect = [
            # output of dev?
            b'WL200: SN(201307374), MD(2018-11-23)\r\n' +
            b'WL Range: 1021.509~1072.505nm(Step: 4654~556)\r\n' +
            b'OK\r\n',
            # output of s?
            f"Step: {expected_start_step}\r\n, ERR: -1, Status: 0x340880".encode() +
            b'OK\r\n',
        ]
        self.instr.open()
        self._transport_mock.reset_mock()

    def tearDown(self):
        self.instr.close()
        qmi.stop()
        logging.getLogger("qmi.instruments.wl_photonics.wltf_n").setLevel(logging.NOTSET)

    def test_wrong_int_value(self):
        """Test a call that uses _ask_int to see that it raises an error at wrong response"""
        self.instr._transport.read_until.side_effect = [b"not an int\r\nNOK\r\n"]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_motor_position()

    def test_wrong_float_value(self):
        """Test a call that uses _ask_float to see that it raises an error at wrong response"""
        self.instr._transport.read_until.side_effect = [b"not a float\r\nNOK\r\n"]

        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_center_wavelength()

    def test_get_idn(self):
        """Test ident. """
        vendor = "WL Photonics"
        model = "WL200"
        serial = "201307374"
        version = "2018-11-23"
        self.instr._transport.read_until.side_effect = [
            b'WL200: SN(201307374), MD(2018-11-23)\r\n' +
            b'WL Range: 1021.509~1072.505nm(Step: 4654~556)\r\n' +
            b'OK\r\n',
        ]
        expected_calls = [
            call().write(b'dev?\r\n'),
            call().read_until(b'OK\r\n', timeout=5.0)
        ]

        ident = self.instr.get_idn()

        self._transport_mock.assert_has_calls(expected_calls)
        self.assertEqual(vendor, ident.vendor)
        self.assertEqual(model, ident.model)
        self.assertEqual(serial, ident.serial)
        self.assertEqual(version, ident.version)

    def test_wrong_idn_response(self):
        """Test ident. raises QMI_InstrumentException by a non-sense response."""
        self.instr._transport.read_until.side_effect = [b"nonsense"]
        expected_calls = [
            call().write(b"dev?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT)
        ]
        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_idn()

        self._transport_mock.assert_has_calls(expected_calls)

    def test_set_center_wavelength(self):
        """Test setting the wavelength in nm."""
        input_wl = 1050
        decimals = 3
        self.instr._transport.read_until.side_effect = [
            f"Set Wavelength: {input_wl:.3f}\r\n".encode() + b"OK\r\n"
        ]
        expected_calls = [
            call().write(f"wl{input_wl:.{decimals}f}\r\n".encode()),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT)
        ]
        # Act
        self.instr.set_center_wavelength(input_wl)
        # Assert
        self._transport_mock.assert_has_calls(expected_calls)

    def test_set_center_wavelength_excepts(self):
        """Test setting the wavelength in nm with values out-of-bounds."""
        input_wl = [int(self.wl_min) - 1, int(self.wl_max) + 1]
        # Act
        for inp in input_wl:
            with self.assertRaises(ValueError):
                self.instr.set_center_wavelength(inp)

        # Assert
        self._transport_mock.assert_not_called()

    def test_get_center_wavelength(self):
        """Test getting the wavelength in nm."""
        wl = (self.wl_max - self.wl_min) / 2.0
        expected_wl = round(wl, 3)  # Output wl appears to be in 3 decimals.
        self.instr._transport.read_until.side_effect = [
            f"Wavelength:{wl:.3f}nm\r\nOK\r\n".encode()
        ]
        expected_calls = [
            call().write(b"wl?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT)
        ]
        # Act
        wavelength = self.instr.get_center_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._transport_mock.assert_has_calls(expected_calls)

    def test_get_minimum_wavelength_in_nm(self):
        """Test getting the minimum wavelength in nm."""
        expected_wl = round(self.wl_min, 4)  # We use 4 decimals for rounding output wl.
        # Act
        wavelength = self.instr.get_minimum_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._transport_mock.assert_not_called()

    def test_get_maximum_wavelength_in_nm(self):
        """Test getting the maximum wavelength in nm."""
        expected_wl = round(self.wl_max, 4)  # We use 4 decimals for rounding output wl.
        # Act
        wavelength = self.instr.get_maximum_wavelength()
        # Assert
        self.assertEqual(expected_wl, wavelength)
        self._transport_mock.assert_not_called()

    def test_reverse_motor(self):
        """Test moving the motor backwards."""
        input_steps = 200
        self.instr._transport.read_until.side_effect = [
            f"SB: {input_steps}\r\nOK\r\n".encode()
        ]
        expected_write_calls = [
            call().write(f"sb{input_steps}\r\n".encode())
        ]
        # Act
        self.instr.reverse_motor(input_steps)
        # Assert
        self._transport_mock.assert_has_calls(expected_write_calls)

    def test_reverse_motor_excepts(self):
        """Test setting the steps with values out-of-bounds."""
        input_step = self.step_max
        # Act
        with self.assertRaises(ValueError):
            self.instr.reverse_motor(input_step)

        # Assert
        self._transport_mock.assert_not_called()

    def test_forward_motor(self):
        """Test moving the motor forward."""
        input_steps = 200
        self.instr._transport.read_until.side_effect = [
            f"SF: {input_steps}\r\nOK\r\n".encode()
        ]
        expected_write_calls = [
            call().write(f"sf{input_steps}\r\n".encode())
        ]
        # Act
        self.instr.forward_motor(input_steps)
        # Assert
        self._transport_mock.assert_has_calls(expected_write_calls)

    def test_forward_motor_excepts(self):
        """Test setting the power_level with values out-of-bounds."""
        input_steps = self.step_max
        # Act
        with self.assertRaises(ValueError):
            self.instr.forward_motor(input_steps)

        # Assert
        self._transport_mock.assert_not_called()

    def test_get_motor_position(self):
        """Test getting the steps."""
        expected_step = (self.step_max - self.step_min) // 2
        self.instr._transport.read_until.side_effect = [
            f"Step: {expected_step}, Err: -1, Status: 0x340880\r\nOK\r\n".encode(),
            f"Step: {-expected_step}, Err: 1, Status: 0x340880\r\nOK\r\n".encode()
        ]
        expected_calls = [
            call().write(b"s?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT),
            call().write(b"s?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT)
        ]
        # Act
        steps_1 = self.instr.get_motor_position()
        steps_2 = self.instr.get_motor_position()  # Negative value, also possible
        # Assert
        self.assertEqual(expected_step, steps_1)
        self.assertEqual(-expected_step, steps_2)
        self._transport_mock.assert_has_calls(expected_calls)

    def test_get_minimum_steps(self):
        """Test getting the minimum steps."""
        expected_steps = self.step_min
        # Act
        steps = self.instr.get_minimum_steps()
        # Assert
        self.assertEqual(expected_steps, steps)
        self._transport_mock.assert_not_called()

    def test_get_maximum_steps(self):
        """Test getting the maximum steps."""
        expected_steps = self.step_max
        # Act
        steps = self.instr.get_maximum_steps()
        # Assert
        self.assertEqual(expected_steps, steps)
        self._transport_mock.assert_not_called()

    def test_go_to_zero(self):
        """Test zeroing motor position command."""
        # Shorten sleep time
        self.instr.ZEROING_WAIT = 0.01
        self.instr._transport.read_until.side_effect = [
            b"Go to Zero\r\nOK\r\n",
            b"Step: 0, Err: 0, Status: 0x340880\r\nOK\r\n",
        ]
        expected_write_calls = [
            call().write(b"z\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT),
            call().write(b"s?\r\n"),
            call().read_until(b'OK\r\n', timeout=WlPhotonics_WltfN.DEFAULT_RESPONSE_TIMEOUT),
        ]
        # Act
        self.instr.go_to_zero()
        # Assert
        self.assertEqual(0, self.instr._motor_position)
        self._transport_mock.assert_has_calls(expected_write_calls)



"""Instrument driver for the Siglent SSA3000X spectrum analyzer."""

import logging
import time
from typing import List, Tuple

import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport


_logger = logging.getLogger(__name__)


class SSA3000X(QMI_Instrument):
    """Instrument driver for the Siglent SSA3000X spectrum analyzer."""

    # trace format constants
    _ASCii = "ASCii"
    _FLOAT = "REAL"
    _TRACE_FORMATS = [_ASCii, _FLOAT]

    _RESP_TERMINATOR = "\n"  # Responses have this unconventional terminator
    _TIMEOUT = 2.0  # Can't find value in datasheet. This is best guess, based on testing.

    _CHANNEL_COUNT = 4

    def __init__(self, context: QMI_Context, name: str, transport_descr: str):
        """ """
        super().__init__(context, name)

        self._transport = create_transport(transport_descr)
        self._scpi = ScpiProtocol(transport=self._transport,
                                  response_terminator=self._RESP_TERMINATOR,
                                  default_timeout=self._TIMEOUT)

    def _ask(self, *args, **kwargs) -> str:
        """Helper method to cleanup up the data from the SSA3000X"""
        return self._scpi.ask(*args, **kwargs).replace('\0', '')

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to instrument")
        self._transport.open()

        time.sleep(self._TIMEOUT)
        self._transport.discard_read()  # discard welcome message

        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to instrument")
        super().close()
        self._transport.close()

    @rpc_method
    def get_id(self) -> str:
        """ Query device identification string.

        Returns:
            Device identification string.
        """
        return self._ask("*IDN?", discard=True)

    @rpc_method
    def get_freq_span(self) -> float:
        """ Get the frequency span.

        Returns:
            Frequency span in Hz.
        """
        resp = self._ask(":FREQ:SPAN?", discard=True)
        try:
            return float(resp)
        except ValueError as err:
            raise QMI_InstrumentException(f"Invalid frequency span, received {resp!r}.") from err

    @rpc_method
    def set_freq_span(self, span_freq: float) -> None:
        """ Sets the frequency span.

        Note:
            Setting the span to 0 Hz puts the analyzer into zero span.

        Parameters:
            span_freq: Frequency span in Hz [0 or 100 to 3.2e9].
        """
        if not (span_freq == 0. or 100. <= span_freq <= 3.2e9):
            raise ValueError(f"Invalid frequency span {span_freq}. Must be 0, or between 100 and 3.2e9.")
        
        self._scpi.write(f":FREQ:SPAN {span_freq/1e9:1.9f} GHz")

    @rpc_method
    def get_freq_center(self) -> float:
        """ Get the center frequency.

        Returns:
            Center frequency in Hz.
        """
        resp = self._ask(":FREQ:CENT?", discard=True)
        try:
            return float(resp)
        except ValueError as err:
            raise QMI_InstrumentException(f"Invalid center frequency, received {resp!r}.") from err

    @rpc_method
    def set_freq_center(self, center_freq: float) -> None:
        """ Sets the center frequency of the spectrum analyzer.

        Parameters:
            center_freq: Center frequency in Hz. If zero span, center frequency can be 0 ~ 3.2 GHz,
                         else 50 Hz ~ 3.199999950 GHz.
        """
        # Frequency span determines range check.
        if self.get_freq_span() == 0:
            if not 0 <= center_freq <= 3.2e9:
                raise ValueError(f"Device has zero span. Center frequency must be between 0 Hz and 3.2 GHz, but is {center_freq}!")
        else:
            if not 50 <= center_freq <= 3.19999995e9:
                raise ValueError(f"Center frequency must be between 50 Hz and 3.199999950 GHz, but is {center_freq}!")

        self._scpi.write(f":FREQ:CENT {center_freq/1e9:1.9f} GHz")

    @rpc_method
    def get_freq_start(self) -> float:
        """ Get the start frequency.

        Returns:
            Start frequency in Hz.
        """
        resp = self._ask(":FREQ:STAR?", discard=True)
        try:
            return float(resp)
        except ValueError as err:
            raise QMI_InstrumentException(f"Invalid start frequency, received {resp!r}.") from err

    @rpc_method
    def get_freq_stop(self) -> float:
        """ Get the stop frequency.

        Returns:
            Stop frequency in Hz.
        """
        resp = self._ask(":FREQ:STOP?", discard=True)
        try:
            return float(resp)
        except ValueError as err:
            raise QMI_InstrumentException(f"Invalid stop frequency, received {resp!r}.") from err

    @rpc_method
    def get_trace_format(self) -> str:
        """ Get the format of the trace data.

        Returns:
            "REAL": single precision floating point values
            "ASCii": ASCii string representation of values
        """
        resp = self._ask(":FORM?", discard=True)
        if resp not in self._TRACE_FORMATS:
            raise QMI_InstrumentException(f"{resp!r} is not a acceptable format {self._TRACE_FORMATS}.")
        return resp

    @rpc_method
    def get_trace(self, channel: int) -> List[float]:
        """ Retrieve trace data from specified channel.

        Parameters:
            Channel for which trace should be retreived.

        Returns:
            The trace data as a list of floats.
        """
        self._validate_channel(channel)
        # Check trace format
        fmt = self.get_trace_format()
        if fmt != self._ASCii:
            raise QMI_InstrumentException(f"Trace format should be {self._ASCii!r}, not {fmt!r}")
        # Read trace data
        resp = self._ask(f":TRAC:DATA? {channel}", discard=True)
        try:
            values_str = resp.split(',')[:-1]  # strip the trailing ','
            return [float(val) for val in values_str]
        except ValueError as err:
            # Raise if either split() or float() fails
            raise QMI_InstrumentException("Trace data could not be coverted to list of floats.") from err

    @rpc_method
    def get_spectrum(self, channel: int) -> Tuple[np.ndarray, np.ndarray]:
        """ Retrieve spectrum data from the specified channel.

        Arguments:
            Channel for which spectrum should be retreived.

        Returns:
            Tuple containing the frequency and amplitude data of the trace of the specified channel.
        """
        self._validate_channel(channel)

        start = self.get_freq_start()
        stop = self.get_freq_stop()
        trace = self.get_trace(channel)

        if start >= stop:
            raise QMI_InstrumentException(f"Start frequency ({start}) is greater or equal to stop frequency ({stop})!")

        if len(trace) == 0:
            raise QMI_InstrumentException("Zero trace data received!")

        ampl = np.array(trace)
        freq = np.linspace(start, stop, len(ampl))

        return freq, ampl

    def _validate_channel(self, channel: int):
        if not 1 <= channel <= self._CHANNEL_COUNT:
            raise QMI_UsageException(f"Channel has to be between 1 and {self._CHANNEL_COUNT}, but is {channel}.")



"""
Instrument driver for the Santec TSL 570 laser.
"""

import logging
from dataclasses import dataclass
from typing import List, Union

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


@dataclass
class _WavelengthRange:
    """Dataclass for wavelength instrument range."""

    min: float
    max: float


@dataclass
class _FrequencyRange:
    """Dataclass for frequency instrument range."""

    min: float
    max: float


@dataclass
class _PowerLevelRange:
    """Dataclass for power level instrument range."""

    min: float
    max: float


class Santec_Tsl570(QMI_Instrument):
    """Instrument driver for the Santec TSL 570 laser.

    The driver is currently based on running at "Legacy" communication, which is also based on
    SCPI protocol, but perhaps not fully complying to it.
    """

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    # power level range in dBm and mW
    POWER_LEVEL_RANGE = {"dBm": (-15.0, 13.0), "mW": (0.04, 20.00)}

    # Valid speeds
    VALID_SPEED_VALUES = [1, 2, 5, 10, 20, 50, 100, 200]

    # System errors
    ERRORS_TABLE = {
        0: "No error",
        -102: "Syntax error",
        -103: "Invalid separator",
        -108: "Parameter not allowed",
        -109: "Missing parameter",
        -113: "Undefined header",
        -148: "Character data not allowed",
        -200: "Execution error",
        -222: "Data out of range",
        -410: "Query INTERRUPTED",
    }

    # System alerts
    ALERTS_TABLE = {
        "No00.": "Power supply Error1",
        "No02.": "Power supply Error2",
        "No03.": "Power supply Error3",
        "No05.": "Wavelength Error",
        "No06.": "Power setting Error",
        "No07.": "Inter lock detection",
        "No20.": "Temperature control Error1",
        "No21.": "Temperature control Error2",
        "No22.": "Temperature control Error3",
        "No23.": "Ongoing Warm up",
        "No25.": "Shutter Error",
        "No26.": "Sensor Error",
        "No27.": "Connection Error",
        "No28.": "Exhaust Fan Error",
    }

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize the instrument driver.

        Parameters:
            context:    The QMI context for running the driver in.
            name:       Name for this instrument instance.
            transport:  QMI transport descriptor to connect to the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport)
        self._scpi_protocol = ScpiProtocol(
            self._transport,
            command_terminator="\r",
            response_terminator="\r",
            default_timeout=self.DEFAULT_RESPONSE_TIMEOUT,
        )
        # Instrument ranges for values
        self._wavelength_range = _WavelengthRange
        self._frequency_range = _FrequencyRange
        self._power_level_range = _PowerLevelRange

    def _check_error(self) -> List[str]:
        """Read the instrument error queue and raise an exception if there is an error.

        Returns:
            errors + alerts: Listed error and alert response strings.
        """
        errors = []
        # When there are no errors, the response is '0,"No error"'.
        while (error := self._scpi_protocol.ask(":SYST:ERR?")) != '0,"No error"':
            errors.append(self.ERRORS_TABLE[int(error.split(",")[0])])

        alerts = []
        # When there are no alerts, the response is "No alerts.".
        while (alert := self._scpi_protocol.ask(":SYST:ALER?")) != "No alerts.":
            alerts.append(self.ALERTS_TABLE[alert.split(",")[0]])

        return errors + alerts

    def _write_and_check_errors(self, cmd: str) -> None:
        """Write a command to the instrument and check if any errors or alerts were raised.

        Parameters:
            cmd: The input command to send.

        Raises:
            QMI_InstrumentException: If there were errors, errors described in the exception message.
        """
        self._scpi_protocol.write(cmd)
        errors = self._check_error()
        if errors:
            _logger.error("[%s] command [%s] resulted in errors: %s", self._name, cmd, errors)
            raise QMI_InstrumentException(f"Command {cmd} resulted in errors: {errors}")

    def _ask_int(self, cmd: str) -> int:
        """Send a query and return a floating point response."""
        resp = self._scpi_protocol.ask(cmd)
        try:
            return int(resp)
        except ValueError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {resp!r}") from exc

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating point response."""
        resp = self._scpi_protocol.ask(cmd)
        try:
            return float(resp)
        except ValueError as exc:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {resp!r}") from exc

    def _ask_bool(self, cmd: str) -> bool:
        """Send a query and return a boolean response."""
        resp = self._scpi_protocol.ask(cmd)
        value = resp.strip().upper()
        if value in ("1", "ON"):
            return True
        elif value in ("0", "OFF"):
            return False
        else:
            raise QMI_InstrumentException(f"Unexpected response to command {cmd!r}: {resp!r}")

    def _set_wavelength_unit(self, terahertz: int) -> None:
        """Set the wavelength unit to terahertz or to nanometers.

        Parameters:
            terahertz: Integer to set wavelength unit to terahertz (1) or to nanometer (0)
        """
        self._write_and_check_errors(f":WAV:UNIT {terahertz}")

    def _set_power_level_unit(self, milliwatts: int) -> None:
        """Set the power_level unit to milliwatts or to decibel-milliwatts. Update the minimum and maximum values
        consequently.

        Parameters:
            milliwatts: Integer to set power level unit to milliwatts (1) or to decibel-milliwatts (0)
        """
        self._write_and_check_errors(f":POW:UNIT {milliwatts}")
        # Then update the minimum and maximum values of the instruments to apply for the changed units.
        unit = "mW" if milliwatts else "dBm"
        self._power_level_range.min = self.POWER_LEVEL_RANGE[unit][0]
        self._power_level_range.max = self.POWER_LEVEL_RANGE[unit][1]

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()
        # Then update the minimum and maximum values of the instrument ranges
        self._wavelength_range.min = self.get_minimum_wavelength()
        self._wavelength_range.max = self.get_maximum_wavelength()
        self._frequency_range.min = self.get_minimum_frequency()
        self._frequency_range.max = self.get_maximum_frequency()
        power_unit = self.get_power_level_unit()
        self._power_level_range.min = self.POWER_LEVEL_RANGE[power_unit][0]
        self._power_level_range.max = self.POWER_LEVEL_RANGE[power_unit][1]

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        self._check_is_open()
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument info and return QMI_InstrumentIdentification instance.

        Returns:
            QMI_InstrumentIdentification: Data with e.g. idn.vendor = SANTEC, idn.model = TSL-570,
            idn.serial = 21020001, idn.version = 0001.000.0001 (firmware version).
        """
        resp = self._scpi_protocol.ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to *IDN?, got {!r}".format(resp))
        return QMI_InstrumentIdentification(
            vendor=words[0].strip(), model=words[1].strip(), serial=words[2].strip(), version=words[3].strip()
        )

    @rpc_method
    def reset(self) -> None:
        """Device reset. Aborts standby operation and clears the command input and error queues."""
        _logger.info("[%s] Resetting instrument", self._name)
        self._write_and_check_errors("*RST")

    @rpc_method
    def clear(self) -> None:
        """Clear Status. Clears all event registers and queues and reflects the summary in the Status Byte Register.
        Clears the following items.
        ・Status Byte Register
        ・Standard Event Status Register
        ・Error Queue
        """
        _logger.info("[%s] Clearing instrument event registers and queues", self._name)
        self._write_and_check_errors("*CLS")

    @rpc_method
    def operation_complete(self) -> bool:
        """Check if the previous operation is completed.

        Returns:
            response: True for yes, False for no.
        """
        response = self._ask_bool("*OPC?")
        return response

    @rpc_method
    def get_errors(self) -> List[str]:
        """Query all errors and alerts.

        Returns:
            errors_and_alerts: A list of all error and alert descriptions.
        """
        return self._check_error()

    @rpc_method
    def set_wavelength(self, wavelength: float) -> None:
        """Set the output wavelength in nanometers.

        Parameters:
            wavelength: The target wavelength in nanometers up to 0.1 pm resolution.

        Raises:
            QMI_InstrumentException: If the wavelength is not in the instrument range.
        """
        unit = "nm"
        dec = 4  # 0.1 pm resolution
        if wavelength < self._wavelength_range.min or wavelength > self._wavelength_range.max:
            raise ValueError(
                f"Wavelength {wavelength:.{dec}f}{unit} out of instrument range "
                f"({self._wavelength_range.min}{unit} - {self._wavelength_range.max}{unit})"
            )

        self._write_and_check_errors(f":WAV {wavelength:.{dec}f}")

    @rpc_method
    def get_wavelength(self) -> float:
        """Get the output wavelength. Default unit is nanometers.

        Returns:
            wavelength: The instrument wavelength in nanometers.
        """
        wavelength = self._ask_float(":WAV?")
        return wavelength

    @rpc_method
    def get_minimum_wavelength(self) -> float:
        """Get the minimum wavelength.

        Returns:
            wavelength: The instrument wavelength minimum in nanometers.
        """
        wavelength = self._ask_float(":WAV:MIN?")
        return wavelength

    @rpc_method
    def get_maximum_wavelength(self) -> float:
        """Get the maximum wavelength.

        Returns:
            wavelength: The instrument wavelength maximum in nanometers.
        """
        wavelength = self._ask_float(":WAV:MAX?")
        return wavelength

    @rpc_method
    def set_frequency(self, frequency: float) -> None:
        """Set the output frequency in terahertz.

        Parameters:
            frequency: The target frequency in terahertz up to 10 MHz resolution.

        Raises:
            QMI_InstrumentException: If the frequency is not in the instrument range.
        """
        unit = "THz"
        dec = 5  # 10 MHz resolution
        if frequency < self._frequency_range.min or frequency > self._frequency_range.max:
            raise ValueError(
                f"frequency {frequency:.{dec}f}{unit} out of instrument range "
                f"({self._frequency_range.min}{unit} - {self._frequency_range.max}{unit})"
            )

        self._write_and_check_errors(f":WAV:FREQ {frequency:.{dec}f}")

    @rpc_method
    def get_frequency(self) -> float:
        """Get the output frequency. Unit depends on unit setting.

        Returns:
            frequency: The instrument frequency in terahertz.
        """
        frequency = self._ask_float(":WAV:FREQ?")
        return frequency

    @rpc_method
    def get_minimum_frequency(self) -> float:
        """Get the minimum frequency.

        Returns:
            frequency: The instrument frequency minimum in terahertz.
        """
        frequency = self._ask_float(":WAV:FREQ:MIN?")
        return frequency

    @rpc_method
    def get_maximum_frequency(self) -> float:
        """Get the maximum frequency.

        Returns:
            frequency: The instrument frequency maximum in terahertz.
        """
        frequency = self._ask_float(":WAV:FREQ:MAX?")
        return frequency

    @rpc_method
    def set_wavelength_unit_to_nm(self) -> None:
        """Set the wavelength unit to nanometers."""
        self._set_wavelength_unit(0)

    @rpc_method
    def set_wavelength_unit_to_thz(self) -> None:
        """Set the wavelength unit to terahertz."""
        self._set_wavelength_unit(1)

    @rpc_method
    def get_wavelength_unit(self) -> str:
        """Get the wavelength unit displayed.

        Returns:
            wl_unit: The wavelength unit as either "THz" or "nm".
        """
        wl_unit = self._ask_bool(":WAV:UNIT?")
        return "THz" if wl_unit else "nm"

    @rpc_method
    def set_wavelength_fine(self, value: float) -> None:
        """Set the wavelength fine-tuning value.

        Parameters:
            value: The fine-tuning value in range -100.00 - +100.00 and maximal resolution of 0.01.

        Raises:
            ValueError: If value is not in the correct range.
        """
        if value < -100.0 or value > 100.0:
            raise ValueError(f"Input value {value} not in valid range [-100...+100].")

        rounded_value = round(value, 2)  # Step resolution is 0.01
        self._write_and_check_errors(f":WAV:FIN {rounded_value:.2f}")

    @rpc_method
    def get_wavelength_fine(self) -> float:
        """Get the wavelength fine-tuning value.

        Returns:
            value: The fine-tuning value in range -100.00 - +100.00 and maximal resolution of 0.01.
        """
        return self._ask_float("WAV:FIN?")

    @rpc_method
    def disable_finetuning_operation(self) -> None:
        """Terminate Fine-Tuning operation."""
        self._write_and_check_errors(":WAV:FIN:DIS")

    @rpc_method
    def set_coherence_control_status(self, status: Union[bool, str]) -> None:
        """Set the Coherence control status.

        Parameters:
            status: New status as either True or "ON", or False or "OFF", to set control ON and OFF, respectively.

        Raises:
            AssertionError: If input parameter is a string but neither "ON" or "OFF".
        """
        if isinstance(status, bool):
            i_status = int(status)

        else:
            assert status.upper() in ("ON", "OFF")
            i_status = 0 if status.upper() == "OFF" else 1

        self._write_and_check_errors(f":COHC {i_status}")

    @rpc_method
    def get_coherence_control_status(self) -> str:
        """Get the Coherence control status.

        Returns:
            status: New status as either "ON", or "OFF".
        """
        status = self._ask_int(":COHC?")
        return "ON" if status else "OFF"

    @rpc_method
    def set_optical_output_status(self, status: Union[bool, str]) -> None:
        """Set the optical output status.

        Parameters:
            status: New status as either True or "ON", or False or "OFF", to set control ON and OFF, respectively.

        Raises:
            AssertionError: If input parameter is a string but neither "ON" or "OFF".
        """
        if isinstance(status, bool):
            i_status = int(status)

        else:
            assert status.upper() in ("ON", "OFF")
            i_status = 0 if status.upper() == "OFF" else 1

        self._write_and_check_errors(f":POW:STAT {i_status}")

    @rpc_method
    def get_optical_output_status(self) -> str:
        """Get the optical output status.

        Returns:
            status: New status as either "ON", or "OFF".
        """
        status = self._ask_int(":POW:STAT?")
        return "ON" if status else "OFF"

    @rpc_method
    def set_power_level(self, power_level: float) -> None:
        """Set optical output power level between -15dBm (~0.03mW) and peak power. Typical peak power from the
        datasheet https://santec.imgix.net/TSL-570-Datasheet.pdf is 13dBm (~20.00mW).

        Parameters:
            power_level: power level in decibel-milliwatts or milliwatts.
        """
        unit = self.get_power_level_unit()
        if power_level < self._power_level_range.min or power_level > self._power_level_range.max:
            raise ValueError(
                f"Power level {power_level:.2f}{unit} out of instrument range "
                f"({self._power_level_range.min}{unit} - {self._power_level_range.max}{unit})"
            )

        self._write_and_check_errors(f":POW {power_level:.2f}")

    @rpc_method
    def get_power_level(self) -> float:
        """Get the output power level setting. Unit depends on unit setting.

        Returns:
            power_level: The instrument power level in decibel-milliwatts or milliwatts.
        """
        power_level = self._ask_float(":POW?")
        return round(power_level, 2)

    @rpc_method
    def get_actual_power(self) -> float:
        """Get the monitored optical power. Unit depends on unit setting.

        Returns:
            power_level: The instrument power level in decibel-milliwatts or milliwatts.
        """
        power_level = self._ask_float(":POW:ACT?")
        return round(power_level, 2)

    @rpc_method
    def get_minimum_power_level(self) -> float:
        """Get the minimum power level.

        Returns:
            power level: The instrument power level minimum in decibel-milliwatts or milliwatts.
        """
        return round(self._power_level_range.min, 2)

    @rpc_method
    def get_maximum_power_level(self) -> float:
        """Get the maximum power level.

        Returns:
            power level: The instrument power level maximum in decibel-milliwatts or milliwatts.
        """
        return round(self._power_level_range.max, 2)

    @rpc_method
    def set_power_level_unit_to_mw(self) -> None:
        """Set the power_level unit to milliwatts."""
        self._set_power_level_unit(1)

    @rpc_method
    def set_power_level_unit_to_dbm(self) -> None:
        """Set the power_level unit to decibel-milliwatts."""
        self._set_power_level_unit(0)

    @rpc_method
    def get_power_level_unit(self) -> str:
        """Get the power_level unit displayed.

        Returns:
            wl_unit: The power_level unit as either "dBm" or "mW".
        """
        wl_unit = self._ask_bool(":POW:UNIT?")
        return "mW" if wl_unit else "dBm"

    @rpc_method
    def set_sweep_start_wavelength(self, wavelength: float) -> None:
        """Set the start wavelength for a sweep.

        Parameters:
            wavelength: wavelength in nanometers.
        """
        unit = "nm"
        dec = 4  # 0.1 pm resolution
        if wavelength < self._wavelength_range.min or wavelength > self._wavelength_range.max:
            raise ValueError(
                f"Wavelength {wavelength:.{dec}f}{unit} out of instrument range "
                f"({self._wavelength_range.min}{unit} - {self._wavelength_range.max}{unit})"
            )

        self._write_and_check_errors(f":WAV:SWE:STAR {wavelength}")

    @rpc_method
    def get_sweep_start_wavelength(self) -> float:
        """Get the sweep start wavelength. Default unit is nanometers.

        Returns:
            wavelength: The instrument wavelength in nanometers.
        """
        wavelength = self._ask_float(":WAV:SWE:STAR?")
        return wavelength

    @rpc_method
    def set_sweep_stop_wavelength(self, wavelength: float) -> None:
        """Set the stop wavelength for a sweep.

        Parameters:
            wavelength: wavelength in nanometers.
        """
        unit = "nm"
        dec = 4  # 0.1 pm resolution
        if wavelength < self._wavelength_range.min or wavelength > self._wavelength_range.max:
            raise ValueError(
                f"Wavelength {wavelength:.{dec}f}{unit} out of instrument range "
                f"({self._wavelength_range.min}{unit} - {self._wavelength_range.max}{unit})"
            )

        self._write_and_check_errors(f":WAV:SWE:STOP {wavelength}")

    @rpc_method
    def get_sweep_stop_wavelength(self) -> float:
        """Get the sweep stop wavelength. Default unit is nanometers.

        Returns:
            wavelength: The instrument wavelength in nanometers.
        """
        wavelength = self._ask_float(":WAV:SWE:STOP?")
        return wavelength

    @rpc_method
    def set_sweep_start_frequency(self, frequency: float) -> None:
        """Set the start frequency for a sweep.

        Parameters:
            frequency: frequency in terahertz.
        """
        unit = "THz"
        dec = 4  # 10 MHz resolution
        if frequency < self._frequency_range.min or frequency > self._frequency_range.max:
            raise ValueError(
                f"Wavelength {frequency:.{dec}f}{unit} out of instrument range "
                f"({self._frequency_range.min}{unit} - {self._frequency_range.max}{unit})"
            )

        self._write_and_check_errors(f":WAV:FREQ:SWE:STAR {frequency}")

    @rpc_method
    def get_sweep_start_frequency(self) -> float:
        """Get the sweep start frequency. Default unit is terahertz.

        Returns:
            frequency: The instrument frequency in nanometers.
        """
        frequency = self._ask_float(":WAV:FREQ:SWE:STAR?")
        return frequency

    @rpc_method
    def set_sweep_stop_frequency(self, frequency: float) -> None:
        """Set the stop frequency for a sweep.

        Parameters:
            frequency: frequency in terahertz.
        """
        unit = "THz"
        dec = 4  # 10 MHz resolution
        if frequency < self._frequency_range.min or frequency > self._frequency_range.max:
            raise ValueError(
                f"Wavelength {frequency:.{dec}f}{unit} out of instrument range "
                f"({self._frequency_range.min}{unit} - {self._frequency_range.max}{unit})"
            )

        self._write_and_check_errors(f":WAV:FREQ:SWE:STOP {frequency}")

    @rpc_method
    def get_sweep_stop_frequency(self) -> float:
        """Get the sweep stop frequency. Default unit is terahertz.

        Returns:
            frequency: The instrument frequency in nanometers.
        """
        frequency = self._ask_float(":WAV:FREQ:SWE:STOP?")
        return frequency

    @rpc_method
    def set_sweep_mode(self, mode: int) -> None:
        """Set sweep mode. Possible modes are:
        0 - Step sweep mode and One way
        1 - Continuous sweep mode and One way
        2 - Step sweep mode and Two way
        3 - Continuous sweep mode and Two way

        Parameters:
            mode: Integer in range [0, 3].
        """
        if mode not in range(4):
            raise ValueError(f"Sweep mode value {mode} not in valid modes range [0-3].")

        self._write_and_check_errors(f":WAV:SWE:MOD {mode}")

    @rpc_method
    def get_sweep_mode(self) -> int:
        """Get sweep mode.

        Returns:
            mode: Integer in range [0, 3].
        """
        return self._ask_int(":WAV:SWE:MOD?")

    @rpc_method
    def set_sweep_speed(self, speed: int) -> None:
        """Set sweep speed. Possible speed values are 1,2,5,10,20,50,100,200nm/s.

        Parameters:
            speed: speed in nm/s.
        """
        if speed not in self.VALID_SPEED_VALUES:
            raise ValueError(f"Sweep speed {speed} not in valid values {self.VALID_SPEED_VALUES}nm/s.")

        self._write_and_check_errors(f":WAV:SWE:SPE {speed}")

    @rpc_method
    def get_sweep_speed(self) -> float:
        """Get sweep speed.

        Returns:
            speed: Speed in range [1, 200]nm/s.
        """
        return self._ask_float(":WAV:SWE:SPE?")

    @rpc_method
    def set_sweep_dwell(self, dwell: float) -> None:
        """Set _stepped_ sweep dwell. Possible dwell range is 0-999.9s in 0.1nm steps. It does not take into account
        delay in one-way sweeps to return to the start frequency.

        Parameters:
            dwell: Dwell between sweep steps in range [0, 1000[s.
        """
        dec = 1
        if dwell < 0.0 or dwell >= 1000.0:
            raise ValueError(f"Sweep dwell {dwell} not in valid range [0-1000[s.")

        self._write_and_check_errors(f":WAV:SWE:DWEL {dwell:.{dec}f}")

    @rpc_method
    def get_sweep_dwell(self) -> float:
        """Get dwell between sweep steps.

        Returns:
            dwell: dwell in range [0, 1000[s.
        """
        return self._ask_float(":WAV:SWE:DWEL?")

    @rpc_method
    def set_sweep_delay(self, delay: float) -> None:
        """Set _continuous_ sweep delay. Possible delay range is 0-999.9s in 0.1nm steps. It does not take into account
        delay in one-way sweeps to return to the start frequency.

        Parameters:
            delay: Delay between sweeps in range [0, 1000[s.
        """
        dec = 1
        if delay < 0.0 or delay >= 1000.0:
            raise ValueError(f"Sweep delay {delay} not in valid range [0-1000[s.")

        self._write_and_check_errors(f":WAV:SWE:DEL {delay:.{dec}f}")

    @rpc_method
    def get_sweep_delay(self) -> float:
        """Get delay between sweeps.

        Returns:
            delay: delay in range [0, 1000[s.
        """
        return self._ask_float(":WAV:SWE:DEL?")

    @rpc_method
    def set_sweep_cycles(self, cycles: int) -> None:
        """Set sweep repetition times. Possible number of repetitions are in range [0, 999].

        Parameters:
            cycles: Integer in range [0, 999].
        """
        if cycles not in range(1000):
            raise ValueError(f"Sweep cycles {cycles} not in valid cycles range [0-999].")

        self._write_and_check_errors(f":WAV:SWE:CYCL {cycles}")

    @rpc_method
    def get_sweep_cycles(self) -> int:
        """Get sweep repetition times.

        Returns:
            cycles: Integer in range [0, 999].
        """
        return self._ask_int(":WAV:SWE:CYCL?")

    @rpc_method
    def set_sweep_state(self, state: int) -> None:
        """Set sweep state. Possible states are:
        0 - Stop
        1 - Start

        Parameters:
            state: Integer in range [0, 1].
        """
        if state not in [0, 1]:
            raise ValueError(f"Sweep state value {state} not in valid states range [0-1].")

        self._write_and_check_errors(f":WAV:SWE:STAT {state}")

    @rpc_method
    def start_repeating_sweep(self) -> None:
        """Start a repeating sweep."""
        self._write_and_check_errors(":WAV:SWE:STAT:REP")

    @rpc_method
    def get_sweep_state(self) -> int:
        """Get sweep state. Possible states are:
        0 - Stopped
        1 - Running
        3 - Standing by trigger
        4 - Preparation for sweep start

        Returns:
            state: Integer in range [0, 4].
        """
        return self._ask_int(":WAV:SWE:STAT?")

    @rpc_method
    def set_trigger_output_timing_mode(self, mode: int) -> None:
        """Set trigger output timing mode. Possible modes are:
        0 - None
        1 - Stop
        2 - Start
        3 - Step

        Parameters:
            mode: Integer in range [0, 3].
        """
        if mode not in range(4):
            raise ValueError(f"Trigger output timing mode {mode} not in valid modes range [0-3].")

        self._write_and_check_errors(f":TRIG:OUTP {mode}")

    @rpc_method
    def get_trigger_output_timing_mode(self) -> int:
        """Get trigger output timing mode.

        Returns:
            mode: Integer in range [0, 3].
        """
        return self._ask_int(":TRIG:OUTP?")

    @rpc_method
    def set_trigger_output_period_mode(self, mode: int) -> None:
        """Set trigger output period mode. Possible modes are:
        0 - Sets the output trigger to be periodic in time.
        1 - Sets the output trigger to be periodic in wavelength.

        Parameters:
            mode: Integer in range [0, 1].
        """
        if mode not in [0, 1]:
            raise ValueError(f"Trigger output period mode {mode} not in valid modes [0-1].")

        self._write_and_check_errors(f":TRIG:OUTP:SETT {mode}")

    @rpc_method
    def get_trigger_output_period_mode(self) -> int:
        """Get trigger output period mode.

        Returns:
            mode: Integer in range [0, 1].
        """
        return self._ask_int(":TRIG:OUTP:SETT?")

    @rpc_method
    def set_trigger_output_step(self, step: float) -> None:
        """Set trigger output step. Possible step range is from 0.0001 to ~ maximum specified wavelength
        range in 0.0001nm steps.

        Parameters:
            step: Interval of the trigger signal output [0.0001, max wavelength] in nanometers.
        """
        dec = 4
        max_wl_range = self.get_sweep_stop_wavelength() - self.get_sweep_start_wavelength()
        if step < 0.0001 or step > max_wl_range:
            raise ValueError(f"Trigger output step {step} not in valid range [0.0001-{max_wl_range}]nm.")

        self._write_and_check_errors(f":TRIG:OUTP:STEP {step:.{dec}f}")

    @rpc_method
    def get_trigger_output_step(self) -> float:
        """Get trigger output step.

        Returns:
            step: step in range [0.0001, max wavelength]nm.
        """
        return self._ask_float(":TRIG:OUTP:STEP?")

    @rpc_method
    def issue_soft_trigger(self) -> None:
        """Issues a soft trigger, executing sweep from trigger standby mode."""
        self._write_and_check_errors(":TRIG:INP:SOFT")

    @rpc_method
    def readout_points(self) -> int:
        """reads the number of logging data available.

        Returns:
            The number of logging data available.
        """
        return int(self._scpi_protocol.ask(":READ:POIN?"))

    @rpc_method
    def readout_data(self) -> List[float]:
        """Read out wavelength logging data and convert it into floating point values. According to the manual
        the data points are returned in units of 0.1pm. Thus, value 0x0040F844 (little Endian order) = 4520000
        corresponds to 452.0000nm.

        Returns:
            data: Data points list converted into nanometers.
        """
        data_binary_size = 4  # In Legacy mode
        self._scpi_protocol.write(":READout:DATa?")
        binary_data = self._scpi_protocol.read_binary_data(read_terminator_flag=False)
        hex_data = [binary_data[p : p + data_binary_size] for p in range(0, len(binary_data), data_binary_size)]

        return [int.from_bytes(h_d, byteorder="little") * 1e-4 for h_d in hex_data]

    @rpc_method
    def shutdown(self) -> None:
        """Shut down the device. Breaks also communication with the device."""
        self._write_and_check_errors(":SPEC:SHUT")

    @rpc_method
    def reboot(self) -> None:
        """Restarts the device. This takes about 60 seconds. Note that the communication to the
        device also resets, so continuing with the same proxy is not possible, but a new one needs
        to be made.
        """
        self._write_and_check_errors(":SPECial:REBoot")



"""
TOPTICA Photonics, digital laser controller.

The qmi.instruments.toptica package provides support for:
- DLC pro.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.toptica.dlc import Toptica_DLC as Toptica_DlcPro



""" Instrument driver for the virtual noisy sine generator.
"""

import time
import math
import random

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method


class NoisySineGenerator(QMI_Instrument):
    """Simulated instrument, useful for testing."""

    def __init__(self, context: QMI_Context, name: str) -> None:
        super().__init__(context, name)
        self.frequency = 2.0
        self.amplitude = 100.0
        self.noise = 1.0

    @rpc_method
    def set_frequency(self, value: float) -> None:
        valid = isinstance(value, float) and math.isfinite(value) and value >= 0.0
        if not valid:
            raise ValueError("Bad value for frequency: {!r}".format(value))
        self.frequency = value

    @rpc_method
    def get_frequency(self) -> float:
        return self.frequency

    @rpc_method
    def set_amplitude(self, value: float) -> None:
        valid = isinstance(value, float) and math.isfinite(value) and value >= 0.0
        if not valid:
            raise ValueError("Bad value for amplitude: {!r}".format(value))
        self.amplitude = value

    @rpc_method
    def get_amplitude(self) -> float:
        return self.amplitude

    @rpc_method
    def set_noise(self, value: float) -> None:
        valid = isinstance(value, float) and math.isfinite(value) and value >= 0.0
        if not valid:
            raise ValueError("Bad value for noise: {!r}".format(value))
        self.noise = value

    @rpc_method
    def get_noise(self) -> float:
        return self.noise

    @rpc_method
    def wait(self, duration: float) -> None:
        if duration < 0.0:
            raise ValueError("Bad value for duration: {!r}".format(duration))
        time.sleep(duration)

    @rpc_method
    def get_sample(self) -> float:
        t = time.time()
        return self.amplitude * math.sin(t * self.frequency * 2 * math.pi) + random.gauss(0.0, self.noise)
